import json
from pandas import DataFrame

from mindsdb.utilities.log import get_log
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb_sql.parser.ast import ASTNode


logger = get_log("main")


class HandlerResponse:
    def __init__(self, resp_type: RESPONSE_TYPE, data_frame: DataFrame = None,
                 query: ASTNode = 0, error_code: int = 0, error_message: str = None) -> None:
        self.resp_type = resp_type
        self.query = query
        self.data_frame = data_frame
        self.error_code = error_code
        self.error_message = error_message

    @property
    def type(self):
        return self.resp_type

    def to_json(self):
        try:
            data = None
            if self.data_frame is not None:
                data = self.data_frame.to_json(orient="split", index=False, date_format="iso")
        except Exception as e:
            logger.error("%s.to_json: error - %s", self.__class__.__name__, e)
            data = None
        return  {"type": self.resp_type,
                 "query": self.query,
                 "data_frame": data,
                 "error_code": self.error_code,
                 "error": self.error_message}

    def __repr__(self):
        return "%s: resp_type=%s, query=%s, data_frame=%s, err_code=%s, error=%s" % (
                self.__class__.__name__,
                self.resp_type,
                self.query,
                self.data_frame,
                self.error_code,
                self.error_message
            )

class HandlerStatusResponse:
    def __init__(self, success: bool = True, error_message: str = None, redirect_url: str = None) -> None:
        self.success = success
        self.error_message = error_message
        self.redirect_url = redirect_url

    def to_json(self):
        data = {"success": self.success, "error": self.error_message}
        if self.redirect_url is not None:
            data['redirect_url'] = self.redirect_url
        return data

    def __repr__(self):
        return f"{self.__class__.__name__}: success={self.success},\
              error={self.error_message},\
              redirect_url={self.redirect_url}"


class ExecutorResponse:
    def __init__(self, resp_type: RESPONSE_TYPE, query: object, error_code: int = 0, error_message: str = None):
        self.resp_type = resp_type
        self.query = query

        self.error_code = error_code
        self.error_message = error_message

    @property
    def type(self):
        return self.resp_type

    def to_json(self):
        return  {"type": self.resp_type,
                 "query": self.query,
                 "error_code": self.error_code,
                 "error": self.error_message}

    def __repr__(self):
        return "%s: resp_type=%s, query=%s, err_code=%s, error=%s" % (
                self.__class__.__name__,
                self.resp_type,
                self.query,
                self.error_code,
                self.error_message,
            )



from typing import Optional
from collections import OrderedDict

import pandas as pd
from elasticsearch import Elasticsearch

from elasticsearch.exceptions import ConnectionError, AuthenticationException

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from es.elastic.sqlalchemy import ESDialect
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class ElasticsearchHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Airtable statements.
    """

    name = 'elasticsearch'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'elasticsearch'

        if ('hosts' not in connection_data) and ('cloud_id' not in connection_data):
            raise Exception("Either the hosts or cloud_id parameter should be provided!")

        optional_parameters = ['hosts', 'cloud_id', 'username', 'password']
        for parameter in optional_parameters:
            if parameter not in connection_data:
                connection_data[parameter] = None

        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return StatusResponse(True)

        try:
            self.connection = Elasticsearch(
                hosts=self.connection_data['hosts'].split(','),
                cloud_id=self.connection_data['cloud_id'],
                basic_auth=(self.connection_data['username'], self.connection_data['password'])
            )
            self.is_connected = True
            return StatusResponse(True)
        except ConnectionError as conn_error:
            log.logger.error(f'Connection error when connecting to Elasticsearch: {conn_error}')
            return StatusResponse(False, error_message=str(conn_error))
        except AuthenticationException as auth_error:
            log.logger.error(f'Authentication error when connecting to Elasticsearch: {auth_error}')
            return StatusResponse(False, error_message=str(auth_error))
        except Exception as e:
            log.logger.error(f'Error connecting to Elasticsearch: {e}')
            return StatusResponse(False, error_message=str(e))

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to Elasticsearch {self.connection_data["hosts"]}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()

        try:
            response = connection.sql.query(body={'query': query})
            records = response['rows']
            columns = response['columns']

            new_records = True
            while new_records:
                try:
                    if response['cursor']:
                        response = connection.sql.query(body={'query': query, 'cursor': response['cursor']})

                        new_records = response['rows']
                        records = records + new_records
                except Exception as e:
                    new_records = False

            if records:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        records,
                        columns=[column['name'] for column in columns]
                    )
                )
        except Exception as e:
            log.logger.error(f'Error running query: {query} on {self.connection_data["hosts"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        renderer = SqlalchemyRender(ESDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of indexes that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = """
            SHOW TABLES
        """
        result = self.native_query(query)
        df = result.data_frame
        df = df.drop(['type', 'type'], axis=1)
        result.data_frame = df.rename(columns={'name': 'table_name'})

        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of indexes returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"""
            DESCRIBE {table_name}
        """
        result = self.native_query(query)
        df = result.data_frame
        df = df.drop('mapping', axis=1)
        result.data_frame = df.rename(columns={'column': 'column_name', 'type': 'data_type'})

        return result


connection_args = OrderedDict(
    hosts={
        'type': ARG_TYPE.STR,
        'description': 'The host name(s) or IP address(es) of the Elasticsearch server(s). If multiple host name(s) or '
                       'IP address(es) exist, they should be separated by commas. This parameter is '
                       'optional, but it should be provided if cloud_id is not.'
    },
    cloud_id={
        'type': ARG_TYPE.STR,
        'description': 'The unique ID to your hosted Elasticsearch cluster on Elasticsearch Service. This parameter is '
                       'optional, but it should be provided if hosts is not.'
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Elasticsearch server. This parameter is optional.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Elasticsearch server. This parameter is '
                       'optional.'
    }
)

connection_args_example = OrderedDict(
    hosts='localhost:9200',
)


import replicate
import pandas as pd
from mindsdb.integrations.libs.base import BaseMLEngine
from typing import Dict, Optional
import os
import types
from mindsdb.utilities.config import Config


class ReplicateHandler(BaseMLEngine):
    name = "replicate"

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'using' not in args:
            raise Exception("Replicate engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args['using']

        if 'model_name' not in args or 'version' not in args:
            raise Exception('Both model_name and version must be provided.')

            # Checking if passed model_name and version  are correct or not
        try:
            replicate.default_client.api_token = args['api_key']
            replicate.models.get(args['model_name']).versions.get(args['version'])

        except Exception as e:
            if e.args[0].startswith('Not found'):
                raise Exception(f"Could not retrieve version {args['version']} of model {args['model_name']}. Verify values are correct and try again.", e)

            elif e.args[0].startswith('Incorrect authentication token'):
                raise Exception("Provided api_key is Incorrect. Get your api_key here: https://replicate.com/account/api-tokens")

            else:
                raise Exception(f"Error occured.", e)

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """Saves model details in storage to access it later
        """
        args = args['using']
        args['target'] = target
        self.model_storage.json_set('args', args)

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        """Using replicate makes the prediction according to your parameters
        
        Args:
            df (pd.DataFrame): The input DataFrame containing data to predict.
            args (Optional[Dict]): Additional arguments for prediction parameters.
        
        Returns:
            pd.DataFrame: The DataFrame containing the predicted results.
        """

        # Extracting prediction parameters from input arguments
        pred_args = args['predict_params'] if args else {}
        args = self.model_storage.json_get('args')
        model_name, version, target_col = args['model_name'], args['version'], args['target']

        # Check if '__mindsdb_row_id' column exists and drop it if present
        if '__mindsdb_row_id' in df.columns:
            df.drop(columns=['__mindsdb_row_id'], inplace=True)

        def get_data(conditions):
            # Run prediction using MindsDB's replicate library
            output = replicate.run(
                f"{args['model_name']}:{args['version']}",
                input={**conditions.to_dict(), **pred_args}         # Unpacking parameters inputted
            ) 
            # Process output based on the model type
            if isinstance(output, types.GeneratorType) and args.get('model_type') == 'LLM':
                output = ''.join(list(output))  # If model_type is LLM, make the stream a string
            elif isinstance(output, types.GeneratorType):
                output = list(output)[-1]  # Getting the final URL if output is a generator of frames URL
            elif isinstance(output, list) and len(output) > 0:
                output = output[-1]  # Returns generated image for controlNet models as it outputs filter and generated image
            return output

        # Check if any wrong parameters are given and raise an exception if necessary
        params_names = set(df.columns) | set(pred_args)
        available_params = self._get_schema(only_keys=True)
        wrong_params = []
        for i in params_names:
            if i not in available_params:
                wrong_params.append(i)

        if wrong_params:
            raise Exception(f"""'{wrong_params}' is/are not supported parameter for this model.
Use DESCRIBE PREDICTOR mindsdb.<model_name>.features; to know about available parameters. OR 
Visit https://replicate.com/{model_name}/versions/{version} to check parameters.""")

        # Set the Replicate API token for communication with the server
        replicate.default_client.api_token = self._get_replicate_api_key(args)

        # Run prediction on the DataFrame rows and format the results into a DataFrame
        data = df.apply(get_data, axis=1)
        data = pd.DataFrame(data)
        data.columns = [target_col]

        return data

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:

        if attribute == "features":
            return self._get_schema()

        else:
            return pd.DataFrame(['features'], columns=['tables'])

    def _get_replicate_api_key(self, args, strict=True):
        """ 
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. REPLICATE_API_KEY env variable
            4. replicate.api_key setting in config.json
        """  # noqa
        # 1
        if 'api_key' in args:
            return args['api_key']
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if 'api_key' in connection_args:
            return connection_args['api_key']
        # 3
        api_key = os.getenv('REPLICATE_API_TOKEN')
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        replicate_cfg = config.get('replicate', {})
        if 'api_key' in replicate_cfg:
            return replicate_cfg['api_key']

        if strict:
            raise Exception(f'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.')

    def _get_schema(self, only_keys=False):
        '''Return parameters list with its description, default value and type,
         which helps user to customize their prediction '''

        args = self.model_storage.json_get('args')
        os.environ['REPLICATE_API_TOKEN'] = self._get_replicate_api_key(args)
        replicate.default_client.api_token = self._get_replicate_api_key(args)
        model = replicate.models.get(args['model_name'])
        version = model.versions.get(args['version'])
        schema = version.get_transformed_schema()['components']['schemas']['Input']['properties']

        # returns only list of paramater
        if only_keys:
            return schema.keys()

        for i in list(schema.keys()):
            for j in list(schema[i].keys()):
                if j not in ['default', 'description', 'type']:
                    schema[i].pop(j)

        df = pd.DataFrame(schema).T
        df = df.reset_index().rename(columns={'index': 'inputs'})
        return df.fillna('-')




class UnsupportedColumnException(Exception):
    """
    Exception raised when a column that is not supported is used in a query.
    """


class MandatoryColumnException(Exception):
    """
    Exception raised when a mandatory column is missing from a query.
    """


class ColumnCountMismatchException(Exception):
    """
    Exception raised when the number of columns in the query does not match the number of values.
    """


import os
import tempfile
from pathlib import Path

import psutil
from flask import request
from flask_restx import Resource
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.util import ns_conf
from mindsdb.utilities.telemetry import (
    enable_telemetry,
    disable_telemetry,
    telemetry_file_exists,
    inject_telemetry_to_static
)
from mindsdb.api.http.gui import update_static
from mindsdb.utilities.fs import clean_unlinked_process_marks
from mindsdb.api.http.utils import http_error


def get_active_tasks():
    response = {
        'learn': False,
        'predict': False,
        'analyse': False
    }

    if os.name != 'posix':
        return response

    for process_type in response:
        processes_dir = Path(tempfile.gettempdir()).joinpath(f'mindsdb/processes/{process_type}/')
        if not processes_dir.is_dir():
            continue
        clean_unlinked_process_marks()
        process_marks = [x.name for x in processes_dir.iterdir()]
        if len(process_marks) > 0:
            response[process_type] = True

    return response


@ns_conf.route('/ping')
class Ping(Resource):
    @ns_conf.doc('get_ping')
    def get(self):
        '''Checks server avaliable'''
        return {'status': 'ok'}


@ns_conf.route('/ping/ml_task_queue')
class PingMLTaskQueue(Resource):
    @ns_conf.doc('get_ping_ml_task_queue')
    def get(self):
        '''Check if ML tasks queue process is alive'''
        processes_dir = Path(tempfile.gettempdir()).joinpath('mindsdb/processes/internal/')
        if processes_dir.is_dir():
            ml_tasks_queue_mark = next((x for x in processes_dir.iterdir() if x.name.endswith('ml_task_consumer')), None)
            if ml_tasks_queue_mark is not None:
                try:
                    pid = int(ml_tasks_queue_mark.name.split('-')[0])
                    process = psutil.Process(pid)
                    if process.status() in (psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD):
                        raise psutil.NoSuchProcess(pid)
                    return '', 200
                except Exception:
                    return '', 404
        return '', 404


@ns_conf.route('/readiness')
class ReadinessProbe(Resource):
    @ns_conf.doc('get_ready')
    def get(self):
        '''Checks server is ready for work'''

        tasks = get_active_tasks()
        for key in tasks:
            if tasks[key] is True:
                return http_error(503, 'not ready', 'not ready')

        return '', 200


@ns_conf.route('/ping_native')
class PingNative(Resource):
    @ns_conf.doc('get_ping_native')
    def get(self):
        ''' Checks server use native for learn or analyse.
            Will return right result only on Linux.
        '''
        return get_active_tasks()


@ns_conf.route('/telemetry')
class Telemetry(Resource):
    @ns_conf.doc('get_telemetry_status')
    def get(self):
        storage_dir = ca.config_obj['storage_dir']
        status = "enabled" if telemetry_file_exists(storage_dir) else "disabled"
        return {"status": status}

    @ns_conf.doc('set_telemetry')
    def post(self):
        data = request.json
        action = data['action']
        if str(action).lower() in ["true", "enable", "on"]:
            enable_telemetry(ca.config_obj['storage_dir'])
        else:
            disable_telemetry(ca.config_obj['storage_dir'])
        inject_telemetry_to_static(ca.config_obj.paths['static'])
        return '', 200


@ns_conf.route('/validate_json_ai')
class ValidateJsonAI(Resource):
    def post(self):
        json_ai = request.json.get('json_ai')
        if json_ai is None:
            return 'Please provide json_ai', 400
        try:
            lw_handler = ca.integration_controller.get_handler('lightwood')
            code = lw_handler.code_from_json_ai(json_ai)
        except Exception as e:
            return {'error': str(e)}
        return {'code': code}


@ns_conf.route('/update-gui')
class UpdateGui(Resource):
    @ns_conf.doc('get_update_gui')
    def get(self):
        update_static()
        return '', 200



from contextlib import closing

import pymssql
import pandas as pd

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


class SqlServerHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Microsoft SQL Server statements. 
    """
    name = 'mssql'

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_args = kwargs
        self.connection_data = self.connection_args.get('connection_data')
        self.dialect = 'mssql'
        self.database = self.connection_data.get('database')

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Handles the connection to a SQL Server insance.
        """
        if self.is_connected is True:
            return self.connection

        self.connection = pymssql.connect(**self.connection_data)
        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the SQL Server database
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to SQL Server {self.database}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in SQL Server
        :return: returns the records from the current recordset
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        # with closing(connection) as con:
        with connection.cursor(as_dict=True) as cur:
            try:
                cur.execute(query)
                result = cur.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                connection.commit()
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.database}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender('mssql')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in MySQL
        """
        query = f"""
            SELECT
                table_schema,
                table_name,
                table_type
            FROM {self.database}.INFORMATION_SCHEMA.TABLES
            WHERE TABLE_TYPE in ('BASE TABLE', 'VIEW');
        """
        result = self.native_query(query)
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"""
            SELECT
                column_name as "Field",
                data_type as "Type"
            FROM
                information_schema.columns
            WHERE
                table_name = '{table_name}'
        """
        result = self.native_query(q)
        return result



from datetime import datetime, timedelta


def make_ssl_cert(file_path):
    from cryptography import x509
    from cryptography.x509.oid import NameOID
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )

    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, 'mdb_autogen'),
        x509.NameAttribute(NameOID.COUNTRY_NAME, 'US'),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, 'California'),
        x509.NameAttribute(NameOID.LOCALITY_NAME, 'Berkeley'),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'MindsDB')
    ])

    now = datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
        .subject_name(name)
        .issuer_name(name)
        .public_key(key.public_key())
        .serial_number(1)
        .not_valid_before(now - timedelta(days=10 * 365))
        .not_valid_after(now + timedelta(days=10 * 365))
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=0),
            False
        )
        .sign(key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
    key_pem = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    with open(file_path, 'wb') as f:
        f.write(key_pem + cert_pem)



import time
import socket
import os
import datetime as dt

import sqlalchemy as sa

from mindsdb.utilities import log
from mindsdb.utilities.log import initialize_log

from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage import db

from .task_thread import TaskThread


class TaskMonitor:

    MONITOR_INTERVAL_SECONDS = 2
    LOCK_EXPIRED_SECONDS = MONITOR_INTERVAL_SECONDS * 30

    def __init__(self):
        self._active_tasks = {}

    def start(self):
        config = Config()
        db.init()
        initialize_log(config, 'jobs', wrap_print=True)
        self.config = config

        while True:
            try:
                self.check_tasks()

                db.session.rollback()  # disable cache
                time.sleep(self.MONITOR_INTERVAL_SECONDS)

            except (SystemExit, KeyboardInterrupt):
                self.stop_all_tasks()
                return

            except Exception as e:
                log.logger.error(e)
                db.session.rollback()

    def stop_all_tasks(self):

        active_tasks = list(self._active_tasks.keys())
        for task_id in active_tasks:
            self.stop_task(task_id)

    def check_tasks(self):
        allowed_tasks = set()

        for task in db.session.query(db.Tasks).filter(db.Tasks.active == True): # noqa
            allowed_tasks.add(task.id)

            # start new tasks
            if task.id not in self._active_tasks:
                self.start_task(task)

        # Check active tasks
        active_tasks = list(self._active_tasks.items())
        for task_id, task in active_tasks:

            if task_id not in allowed_tasks:
                # old task
                self.stop_task(task_id)

            elif not task.is_alive():
                # dead task
                self.stop_task(task_id)

            else:
                # need to be reloaded ?
                record = db.Tasks.query.get(task_id)
                if record.reload:
                    record.reload = False
                    self.stop_task(task_id)
                else:
                    # set alive time of running tasks
                    self._set_alive(task_id)

    def _lock_task(self, task):
        run_by = f'{socket.gethostname()} {os.getpid()}'
        db_date = db.session.query(sa.func.current_timestamp()).first()[0]
        if task.run_by == run_by:
            # already locked
            task.alive_time = db_date

        elif task.alive_time is None:
            # not locked yet
            task.run_by = run_by
            task.alive_time = db_date

        elif db_date - task.alive_time > dt.timedelta(seconds=self.LOCK_EXPIRED_SECONDS):
            # lock expired
            task.run_by = run_by
            task.alive_time = db_date

        else:
            return False

        db.session.commit()
        return True

    def _set_alive(self, task_id):
        db_date = db.session.query(sa.func.current_timestamp()).first()[0]
        task = db.Tasks.query.get(task_id)
        task.alive_time = db_date
        db.session.commit()

    def _unlock_task(self, task_id):
        task = db.Tasks.query.get(task_id)
        if task is not None:
            task.alive_time = None
            db.session.commit()

    def start_task(self, task):
        if not self._lock_task(task):
            # can't lock, skip
            return

        thread = TaskThread(task.id)

        thread.start()
        self._active_tasks[task.id] = thread

    def stop_task(self, task_id: int):
        thread = self._active_tasks[task_id]
        thread.stop()
        thread.join(1)

        if thread.is_alive():
            # don't delete task, wait next circle
            return

        del self._active_tasks[task_id]
        self._unlock_task(task_id)


def start(verbose=False):

    monitor = TaskMonitor()
    monitor.start()


if __name__ == '__main__':
    start()



import pandas as pd

from mindsdb.integrations.handlers.rag_handler.ingest import RAGIngestor
from mindsdb.integrations.handlers.writer_handler.settings import (
    WriterHandlerParameters,
)


class WriterIngestor(RAGIngestor):
    def __init__(self, args: WriterHandlerParameters, df: pd.DataFrame):
        super().__init__(args, df)



import json
import traceback
from http import HTTPStatus

from flask import request
from flask_restx import Resource

from mindsdb.interfaces.storage.fs import FileStorageFactory, RESOURCE_GROUP
from mindsdb.api.http.namespaces.configs.tabs import ns_conf
from mindsdb.utilities.log import get_log
from mindsdb.utilities.context import context as ctx
from mindsdb.api.http.utils import http_error

logger = get_log("main")


TABS_FILENAME = 'tabs'


def get_storage():

    storageFactory = FileStorageFactory(
        resource_group=RESOURCE_GROUP.TAB,
        sync=True
    )

    # resource_id is useless for 'tabs'
    # use constant
    return storageFactory(0)


@ns_conf.route('/')
class Tab(Resource):
    @ns_conf.doc('get_tabs')
    def get(self):
        company_id = request.headers.get("company-id", None)
        ctx.company_id = company_id
        storage = get_storage()
        tabs = None
        try:
            raw_data = storage.file_get(TABS_FILENAME)
            tabs = json.loads(raw_data)
        except Exception as e:
            logger.warning("unable to get tabs data - %s", e)
            return {}, 200
        return tabs, 200

    @ns_conf.doc('save_tabs')
    def post(self):
        company_id = request.headers.get("company-id", None)
        ctx.company_id = company_id
        storage = get_storage()
        try:
            tabs = request.json
            b_types = json.dumps(tabs).encode("utf-8")
            storage.file_set(TABS_FILENAME, b_types)
        except Exception as e:
            logger.warning("unable to store tabs data - %s", e)
            print(traceback.format_exc())
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                "Can't save tabs",
                'something went wrong during tabs saving'
            )

        return '', 200



import unittest
from mindsdb.integrations.handlers.lindorm_handler.lindorm_handler import LindormHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class LindormHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "url": 'http://127.0.0.1:8765',
            "autocommit": True
        }
        cls.handler = LindormHandler('test_phoenix_handler', cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_native_query_select(self):
        query = "SELECT * FROM USERS"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_3_get_columns(self):
        columns = self.handler.get_columns('USERS')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



import mindsdb_sql
from mindsdb_sql import parse_sql
from mindsdb_sql.planner import utils as planner_utils

from mindsdb.api.mysql.mysql_proxy.classes.sql_query import Column, SQLQuery
from mindsdb.api.mysql.mysql_proxy.utilities import (
    ErBadDbError,
    SqlApiException,
    logger,
)
import mindsdb.utilities.profiler as profiler
from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands


class Executor:
    """This class stores initial and intermediate params
    between different steps of query execution. And it is also
    creates a separate instance of ExecuteCommands to execute the current
    query step.

    IMPORTANT: A public API of this class is a contract.
    And there are at least 2 classes strongly depend on it:
        ExecuctorClient
        ExecutorService.
    These classes do the same work as Executor when
    MindsDB works in 'modularity' mode.
    Thus please make sure that IF you change the API,
    you must update the API of these two classes as well!"""

    def __init__(self, session, sqlserver):
        self.session = session
        self.sqlserver = sqlserver

        self.query = None

        # returned values
        # all this attributes needs to be added in
        # self.json() method
        self.columns = []
        self.params = []
        self.data = None
        self.state_track = None
        self.server_status = None
        self.is_executed = False
        self.error_message = None
        self.error_code = None

        # self.predictor_metadata = {}

        self.sql = ""
        self.sql_lower = ""

        self.command_executor = ExecuteCommands(self.session, self)

    def change_default_db(self, new_db):
        self.command_executor.change_default_db(new_db)

    def stmt_prepare(self, sql):

        resp = self.execute_external(sql)
        if resp is not None:
            # is already executed
            self.is_executed = True
            return

        self.parse(sql)

        # if not params
        params = planner_utils.get_query_params(self.query)
        if len(params) == 0:
            # execute immediately
            self.do_execute()

        else:
            # plan query
            # TODO less complex.
            #  planner is inside SQLQuery now.

            sqlquery = SQLQuery(self.query, session=self.session, execute=False)

            sqlquery.prepare_query()

            self.params = [
                Column(
                    alias=p.value,
                    type="str",
                    name=p.value,
                )
                for p in params
            ]

            # TODO:
            #   select * from mindsdb.models doesn't invoke prepare_steps and columns_list is empty
            self.columns = sqlquery.columns_list

    def stmt_execute(self, param_values):
        if self.is_executed:
            return

        # fill params
        self.query = planner_utils.fill_query_params(self.query, param_values)

        # execute query
        self.do_execute()

    @profiler.profile()
    def query_execute(self, sql):
        logger.info("%s.query_execute: sql - %s", self.__class__.__name__, sql)
        resp = self.execute_external(sql)
        if resp is not None:
            # is already executed
            self.is_executed = True
            return

        self.parse(sql)
        self.do_execute()

    # for awesome Mongo API only
    def binary_query_execute(self, sql):
        self.sql = sql.to_string()
        self.sql_lower = self.sql.lower()

        ret = self.command_executor.execute_command(sql)
        self.error_code = ret.error_code
        self.error_message = ret.error_message

        self.data = ret.data
        self.server_status = ret.status
        if ret.columns is not None:
            self.columns = ret.columns

        self.state_track = ret.state_track

    def execute_external(self, sql):

        # not exec directly in integration
        return None

        # try exec in external integration
        if (
            isinstance(self.session.database, str)
            and len(self.session.database) > 0
            and self.session.database.lower()
            not in ("mindsdb", "files", "information_schema")
            and "@@" not in sql.lower()
            and (
                (sql.lower().strip().startswith("select") and "from" in sql.lower())
                or (
                    sql.lower().strip().startswith("show")
                    # and 'databases' in sql.lower()
                    and "tables" in sql.lower()
                )
            )
        ):
            datanode = self.session.datahub.get(self.session.database)
            if datanode is None:
                raise ErBadDbError("Unknown database - %s" % self.session.database)

            # try parse or send raw sql
            try:
                sql = parse_sql(sql, dialect="mindsdb")
            except mindsdb_sql.exceptions.ParsingException:
                pass

            result, column_info = datanode.query(sql)
            columns = [
                Column(name=col["name"], type=col["type"]) for col in column_info
            ]

            data = []
            if len(result) > 0:
                # columns = [{
                #     'table_name': '',
                #     'name': x,
                #     'type': TYPES.MYSQL_TYPE_VAR_STRING
                # } for x in result[0].keys()]

                data = [[str(value) for key, value in x.items()] for x in result]
            self.columns = columns
            self.data = data
            return True

    @profiler.profile()
    def parse(self, sql):
        logger.info("%s.parse: sql - %s", self.__class__.__name__, sql)
        self.sql = sql
        sql_lower = sql.lower()
        self.sql_lower = sql_lower.replace("`", "")

        try:
            self.query = parse_sql(sql, dialect="mindsdb")
        except Exception as mdb_error:
            try:
                self.query = parse_sql(sql, dialect="mysql")
            except Exception:
                # not all statements are parsed by parse_sql
                logger.warning(f"SQL statement is not parsed by mindsdb_sql: {sql}")

                raise SqlApiException(
                    f"SQL statement cannot be parsed by mindsdb_sql - {sql}: {mdb_error}"
                ) from mdb_error

                # == a place for workarounds ==
                # or run sql in integration without parsing

    @profiler.profile()
    def do_execute(self):
        # it can be already run at prepare state
        logger.info("%s.do_execute", self.__class__.__name__)
        if self.is_executed:
            return

        ret = self.command_executor.execute_command(self.query)
        self.error_code = ret.error_code
        self.error_message = ret.error_message

        self.is_executed = True

        self.data = ret.data
        self.server_status = ret.status
        if ret.columns is not None:
            self.columns = ret.columns

        self.state_track = ret.state_track



import os
import json
import openai
import contextlib
from typing import Optional, Dict

import pandas as pd

from mindsdb.integrations.handlers.openai_handler.openai_handler import OpenAIHandler
from mindsdb.integrations.handlers.openai_handler.constants import OPENAI_API_BASE


# TODO: retrieve these from API once possible
CHAT_MODELS = (
    'meta-llama/Llama-2-7b-chat-hf',
    'meta-llama/Llama-2-13b-chat-hf',
    'meta-llama/Llama-2-70b-chat-hf',
    'codellama/CodeLlama-34b-Instruct-hf',
)
ANYSCALE_API_BASE = 'https://api.endpoints.anyscale.com/v1'


class AnyscaleEndpointsHandler(OpenAIHandler):
    name = 'anyscale_endpoints'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.all_models = list(CHAT_MODELS)
        self.chat_completion_models = CHAT_MODELS
        self.supported_ft_models = CHAT_MODELS  # base models compatible with fine-tuning
        self.default_model = CHAT_MODELS[0]
        self.base_api = ANYSCALE_API_BASE
        self.default_mode = 'default'  # can also be 'conversational' or 'conversational-full'
        self.supported_modes = ['default', 'conversational', 'conversational-full']
        self.rate_limit = 25  # requests per minute
        self.max_batch_size = 20
        self.default_max_tokens = 100
        self.ft_cls = openai.FineTuningJob  # non-legacy fine-tuning endpoint

    @staticmethod
    @contextlib.contextmanager
    def _anyscale_base_api(key='OPENAI_API_BASE'):
        """ Temporarily updates the API base env var to point towards the Anyscale URL. """
        old_base = os.environ.get(key, OPENAI_API_BASE)
        os.environ[key] = ANYSCALE_API_BASE
        try:
            yield  # enter
        finally:
            os.environ[key] = old_base  # exit

    def create(self, target, args=None, **kwargs):
        with self._anyscale_base_api():
            # load fine-tuned models, then hand over
            _args = self.model_storage.json_get('args')
            base_models = self.chat_completion_models
            self.chat_completion_models = _args.get('chat_completion_models', base_models) if _args else base_models
            super().create(target, args, **kwargs)

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        with self._anyscale_base_api():
            # load fine-tuned models, then hand over
            _args = self.model_storage.json_get('args')
            base_models = self.chat_completion_models
            self.chat_completion_models = _args.get('chat_completion_models', base_models) if _args else base_models
            return super().predict(df, args)

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        with self._anyscale_base_api():
            super().finetune(df, args)
            # rewrite chat_completion_models to include the newly fine-tuned model
            args = self.model_storage.json_get('args')
            args['chat_completion_models'] = list(self.chat_completion_models) + [args['model_name']]
            self.model_storage.json_set('args', args)

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        if 'api_key' in args:
            del args['api_key']

        if attribute == 'args':
            return pd.DataFrame(args.items(), columns=['key', 'value'])
        elif attribute == 'metadata':
            # we opt for the URL because some models require completing a form to access their artifacts
            model_name = args.get('model_name', self.default_model)
            model_card_url = 'https://huggingface.co/' + model_name
            return pd.DataFrame({'model_name': [model_name], 'model_card': [model_card_url]})
        else:
            tables = ['args', 'metadata']
            return pd.DataFrame(tables, columns=['tables'])

    @staticmethod
    def _check_ft_cols(df, cols):
        for col in ['role', 'content']:
            if col not in set(df.columns):
                raise Exception(f"To fine-tune this model, format your select data query to have a `role` column and a `content` column.")  # noqa

    def _prepare_ft_jsonl(self, df, temp_storage_path, temp_filename, _, test_size=0.2):
        """
            df: has exactly two columns, `role` and `content`. Rows contain >= 1 chats in long (stacked) format.
            For more details, check `FineTuning -> Data Format` in the Anyscale API reference.
        """
        # 1. aggregate each chat sequence into one row
        chats = []
        chat = []
        for i, row in df.iterrows():
            if row['role'] == 'system' and len(chat) > 0:
                chats.append({'messages': chat})
                chat = []
            event = {'role': row['role'], 'content': row['content']}
            chat.append(event)
        chats.append({'messages': chat})
        series = pd.Series(chats)
        train = series.iloc[:int(len(series) * (1 - test_size))]
        val = series.iloc[-int(len(series) * test_size) - 1:]

        # 2. write as jsonl
        file_names = {
            'train': f'{temp_filename}_prepared_train.jsonl',
            'val': f'{temp_filename}_prepared_valid.jsonl',
        }
        train.to_json(os.path.join(temp_storage_path, file_names['train']), orient='records', lines=True)
        val.to_json(os.path.join(temp_storage_path, file_names['val']), orient='records', lines=True)

        # 3. validate
        self._validate_jsonl(os.path.join(temp_storage_path, file_names['train']))
        self._validate_jsonl(os.path.join(temp_storage_path, file_names['val']))
        return file_names

    @staticmethod
    def _get_ft_model_type(model_name: str):
        for base_model in CHAT_MODELS:
            if base_model.lower() in model_name.lower():
                return base_model
        raise Exception(f'Model {model_name} cannot be finetuned.')

    @staticmethod
    def _add_extra_ft_params(ft_params, using_args):
        hyperparameters = {}
        # we populate separately because keys with `None` break the API
        for key in ('n_epochs', 'context_length'):
            if using_args.get(key, None):
                hyperparameters[key] = using_args[key]
        if hyperparameters:
            return {**ft_params, **{'hyperparameters': hyperparameters}}
        else:
            return ft_params

    @staticmethod
    def _validate_jsonl(jsonl_path):
        """ Borrowed from Anyscale docs. We may want something customized in the future though. """
        with open(jsonl_path, 'r', encoding='utf-8') as f:
            items = [json.loads(line) for line in f]

        def check_data_for_format_errors(items: list):
            for line_num, batch in enumerate(items):
                prefix = f"Error in line #{line_num + 1}: "
                if not isinstance(batch, dict):
                    raise Exception(f"{prefix}Each line in the provided data should be a dictionary")

                if "messages" not in batch:
                    raise Exception(f"{prefix}Each line in the provided data should have a 'messages' key")

                if not isinstance(batch["messages"], list):
                    raise Exception(f"{prefix}Each line in the provided data should have a 'messages' key with a list of messages")  # noqa

                messages = batch["messages"]
                if not any(message.get("role", None) == "assistant" for message in messages):
                    raise Exception(f"{prefix}Each message list should have at least one message with role 'assistant'")  # noqa

                for message_num, message in enumerate(messages):
                    prefix = f"Error in line #{line_num + 1}, message #{message_num + 1}: "
                    if "role" not in message or "content" not in message:
                        raise Exception(f"{prefix}Each message should have a 'role' and 'content' key")

                    if any(k not in ("role", "content", "name") for k in message):
                        raise Exception(f"{prefix}Each message should only have 'role', 'content', and 'name' keys, any other key is not allowed")  # noqa

                    if message.get("role", None) not in ("system", "user", "assistant"):
                        raise Exception(f"{prefix}Each message should have a valid role (system, user, or assistant)")  # noqa

        try:
            check_data_for_format_errors(items)
        except Exception as e:
            raise Exception(f"Fine-tuning data format is not valid. Got: {e}")



import os
import importlib
from pathlib import Path
import tempfile
import multipart

from flask import request, send_file, abort, current_app as ca
from flask_restx import Resource

from mindsdb_sql.parser.ast import Identifier
from mindsdb_sql.parser.dialects.mindsdb import CreateMLEngine

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.handlers import ns_conf
from mindsdb.integrations.utilities.install import install_dependencies

from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands


@ns_conf.route('/')
class HandlersList(Resource):
    @ns_conf.doc('handlers_list')
    def get(self):
        '''List all db handlers'''
        handlers = ca.integration_controller.get_handlers_import_status()
        result = []
        for handler_type, handler_meta in handlers.items():
            row = {'name': handler_type}
            row.update(handler_meta)
            result.append(row)
        return result


@ns_conf.route('/<handler_name>/icon')
class HandlerIcon(Resource):
    @ns_conf.param('handler_name', 'Handler name')
    def get(self, handler_name):
        try:
            handlers_import_status = ca.integration_controller.get_handlers_import_status()
            icon_name = handlers_import_status[handler_name]['icon']['name']
            handler_folder = handlers_import_status[handler_name]['import']['folder']
            mindsdb_path = Path(importlib.util.find_spec('mindsdb').origin).parent
            icon_path = mindsdb_path.joinpath('integrations/handlers').joinpath(handler_folder).joinpath(icon_name)
            if icon_path.is_absolute() is False:
                icon_path = Path(os.getcwd()).joinpath(icon_path)
        except Exception:
            return abort(404)
        else:
            return send_file(icon_path)


@ns_conf.route('/<handler_name>/install')
class InstallDependencies(Resource):
    @ns_conf.param('handler_name', 'Handler name')
    def post(self, handler_name):
        handler_import_status = ca.integration_controller.get_handlers_import_status()
        if handler_name not in handler_import_status:
            return f'Unkown handler: {handler_name}', 400

        if handler_import_status[handler_name].get('import', {}).get('success', False) is True:
            return 'Installed', 200

        handler_meta = handler_import_status[handler_name]

        dependencies = handler_meta['import']['dependencies']
        if len(dependencies) == 0:
            return 'Installed', 200

        result = install_dependencies(dependencies)

        # reload it if any result, so we can get new error message
        ca.integration_controller.reload_handler_module(handler_name)
        if result.get('success') is True:
            return '', 200
        return http_error(
            500,
            'Failed to install dependency',
            result.get('error_message', 'unknown error')
        )


@ns_conf.route('/byom/<name>')
@ns_conf.param('name', "Name of the model")
class BYOMUpload(Resource):
    @ns_conf.doc('put_file')
    def put(self, name):
        ''' upload new model
            params in FormData:
                - code
                - modules
        '''

        params = {}

        def on_field(field):
            name = field.field_name.decode()
            value = field.value.decode()
            params[name] = value

        def on_file(file):
            params[file.field_name.decode()] = file.file_object

        temp_dir_path = tempfile.mkdtemp(prefix='mindsdb_file_')

        parser = multipart.create_form_parser(
            headers=request.headers,
            on_field=on_field,
            on_file=on_file,
            config={
                'UPLOAD_DIR': temp_dir_path.encode(),  # bytes required
                'UPLOAD_KEEP_FILENAME': True,
                'UPLOAD_KEEP_EXTENSIONS': True,
                'MAX_MEMORY_FILE_SIZE': 0
            }
        )

        while True:
            chunk = request.stream.read(8192)
            if not chunk:
                break
            parser.write(chunk)
        parser.finalize()
        parser.close()

        params['code'].close()
        params['modules'].close()

        sql_session = SessionController()

        command_executor = ExecuteCommands(sql_session, executor=None)

        connection_args = {
            'code': params['code'].name.decode(),
            'modules': params['modules'].name.decode(),
        }

        ast_query = CreateMLEngine(
            name=Identifier(name),
            handler='byom',
            params=connection_args
        )
        command_executor.execute_command(ast_query)

        return '', 200



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: executor.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x65xecutor.proto\"j\n\x0f\x45xecutorContext\x12\n\n\x02id\x18\x01 \x01(\t\x12\x15\n\rconnection_id\x18\x02 \x01(\x05\x12\x12\n\nsession_id\x18\x03 \x01(\t\x12\x0f\n\x07session\x18\x04 \x01(\t\x12\x0f\n\x07\x63ontext\x18\x05 \x01(\t\"B\n\x10\x45xecutionContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0b\n\x03sql\x18\x02 \x01(\t\"H\n\x16\x42inaryExecutionContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0b\n\x03sql\x18\x02 \x01(\x0c\"L\n\x17StatementExecuteContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0e\n\x06params\x18\x02 \x01(\t\"E\n\x10\x44\x65\x66\x61ultDBContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0e\n\x06new_db\x18\x02 \x01(\t\"\xbe\x01\n\x10\x45xecutorResponse\x12\x0f\n\x07\x63olumns\x18\x01 \x01(\x0c\x12\x0e\n\x06params\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x13\n\x0bstate_track\x18\x04 \x01(\t\x12\x15\n\rserver_status\x18\x05 \x01(\t\x12\x13\n\x0bis_executed\x18\x06 \x01(\x08\x12\x0f\n\x07session\x18\x07 \x01(\t\x12\x12\n\nerror_code\x18\x08 \x01(\x05\x12\x15\n\rerror_message\x18\t \x01(\t\"@\n\x16\x45xecutorStatusResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t2\xa6\x04\n\x0f\x45xecutorService\x12:\n\x10StatementPrepare\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x41\n\x10StatementExecute\x12\x18.StatementExecuteContext\x1a\x11.ExecutorResponse\"\x00\x12\x36\n\x0cQueryExecute\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x42\n\x12\x42inaryQueryExecute\x12\x17.BinaryExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x39\n\x0f\x45xecuteExternal\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12/\n\x05Parse\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x32\n\tDoExecute\x12\x10.ExecutorContext\x1a\x11.ExecutorResponse\"\x00\x12\x39\n\x0f\x43hangeDefaultDB\x12\x11.DefaultDBContext\x1a\x11.ExecutorResponse\"\x00\x12=\n\x0e\x44\x65leteExecutor\x12\x10.ExecutorContext\x1a\x17.ExecutorStatusResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'executor_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EXECUTORCONTEXT._serialized_start=18
  _EXECUTORCONTEXT._serialized_end=124
  _EXECUTIONCONTEXT._serialized_start=126
  _EXECUTIONCONTEXT._serialized_end=192
  _BINARYEXECUTIONCONTEXT._serialized_start=194
  _BINARYEXECUTIONCONTEXT._serialized_end=266
  _STATEMENTEXECUTECONTEXT._serialized_start=268
  _STATEMENTEXECUTECONTEXT._serialized_end=344
  _DEFAULTDBCONTEXT._serialized_start=346
  _DEFAULTDBCONTEXT._serialized_end=415
  _EXECUTORRESPONSE._serialized_start=418
  _EXECUTORRESPONSE._serialized_end=608
  _EXECUTORSTATUSRESPONSE._serialized_start=610
  _EXECUTORSTATUSRESPONSE._serialized_end=674
  _EXECUTORSERVICE._serialized_start=677
  _EXECUTORSERVICE._serialized_end=1227
# @@protoc_insertion_point(module_scope)



import os
import sys
import base64
import shutil
import tempfile
import importlib
import threading
import inspect
import multiprocessing
from time import time
from pathlib import Path
from copy import deepcopy
from typing import Optional
from textwrap import dedent
from collections import OrderedDict

from sqlalchemy import func

from mindsdb.interfaces.storage import db
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.fs import FsStore, FileStorage, RESOURCE_GROUP
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE, HANDLER_TYPE
from mindsdb.integrations.handlers_client.db_client_factory import DBClient
from mindsdb.interfaces.model.functions import get_model_records
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.log import get_log
from mindsdb.integrations.libs.ml_exec_base import BaseMLEngineExec
import mindsdb.utilities.profiler as profiler

logger = get_log()


class HandlersCache:
    """ Cache for data handlers that keep connections opened during ttl time from handler last use
    """

    def __init__(self, ttl: int = 60):
        """ init cache

            Args:
                ttl (int): time to live (in seconds) for record in cache
        """
        self.ttl = ttl
        self.handlers = {}
        self._lock = threading.RLock()
        self._stop_event = threading.Event()
        self.cleaner_thread = None

    def __del__(self):
        self._stop_clean()

    def _start_clean(self) -> None:
        """ start worker that close connections after ttl expired
        """
        if (
            isinstance(self.cleaner_thread, threading.Thread)
            and self.cleaner_thread.is_alive()
        ):
            return
        self._stop_event.clear()
        self.cleaner_thread = threading.Thread(target=self._clean)
        self.cleaner_thread.daemon = True
        self.cleaner_thread.start()

    def _stop_clean(self) -> None:
        """ stop clean worker
        """
        self._stop_event.set()

    def set(self, handler: DatabaseHandler):
        """ add (or replace) handler in cache

            Args:
                handler (DatabaseHandler)
        """
        # do not cache connections in handlers processes
        if multiprocessing.current_process().name.startswith('HandlerProcess'):
            return
        with self._lock:
            try:
                key = (handler.name, ctx.company_id, threading.get_native_id())
                handler.connect()
                self.handlers[key] = {
                    'handler': handler,
                    'expired_at': time() + self.ttl
                }
            except Exception:
                pass
            self._start_clean()

    def get(self, name: str) -> Optional[DatabaseHandler]:
        """ get handler from cache by name

            Args:
                name (str): handler name

            Returns:
                DatabaseHandler
        """
        with self._lock:
            key = (name, ctx.company_id, threading.get_native_id())
            if (
                key not in self.handlers
                or self.handlers[key]['expired_at'] < time()
            ):
                return None
            self.handlers[key]['expired_at'] = time() + self.ttl
            return self.handlers[key]['handler']

    def delete(self, name: str) -> None:
        """ delete handler from cache

            Args:
                name (str): handler name
        """
        with self._lock:
            key = (name, ctx.company_id, threading.get_native_id())
            if key in self.handlers:
                try:
                    self.handlers[key].disconnect()
                except Exception:
                    pass
                del self.handlers[key]
            if len(self.handlers) == 0:
                self._stop_clean()

    def _clean(self) -> None:
        """ worker that delete from cache handlers that was not in use for ttl
        """
        while self._stop_event.wait(timeout=3) is False:
            with self._lock:
                for key in list(self.handlers.keys()):
                    if (
                        self.handlers[key]['expired_at'] < time()
                        and sys.getrefcount(self.handlers[key]) == 2    # returned ref count is always 1 higher
                    ):
                        try:
                            self.handlers[key].disconnect()
                        except Exception:
                            pass
                        del self.handlers[key]
                if len(self.handlers) == 0:
                    self._stop_event.set()


class IntegrationController:
    @staticmethod
    def _is_not_empty_str(s):
        return isinstance(s, str) and len(s) > 0

    def __init__(self):
        self._load_handler_modules()
        self.handlers_cache = HandlersCache()

    def _add_integration_record(self, name, engine, connection_args):
        integration_record = db.Integration(
            name=name,
            engine=engine,
            data=connection_args or {},
            company_id=ctx.company_id
        )
        db.session.add(integration_record)
        db.session.commit()
        return integration_record.id

    def add(self, name, engine, connection_args):

        logger.debug(
            "%s: add method calling name=%s, engine=%s, connection_args=%s, company_id=%s",
            self.__class__.__name__, name, engine, connection_args, ctx.company_id
        )
        handlers_meta = self.get_handlers_import_status()
        handler_meta = handlers_meta[engine]
        accept_connection_args = handler_meta.get('connection_args')
        logger.debug("%s: accept_connection_args - %s", self.__class__.__name__, accept_connection_args)

        files_dir = None
        if accept_connection_args is not None and connection_args is not None:
            for arg_name, arg_value in connection_args.items():
                if (
                    arg_name in accept_connection_args
                    and accept_connection_args[arg_name]['type'] == ARG_TYPE.PATH
                ):
                    if arg_value is None or arg_value == '':
                        continue
                    if files_dir is None:
                        files_dir = tempfile.mkdtemp(prefix='mindsdb_files_')
                    shutil.copy(arg_value, files_dir)
                    connection_args[arg_name] = Path(arg_value).name

        integration_id = self._add_integration_record(name, engine, connection_args)

        if files_dir is not None:
            store = FileStorage(
                resource_group=RESOURCE_GROUP.INTEGRATION,
                resource_id=integration_id,
                sync=False
            )
            store.add(files_dir, '')
            store.push()

        return integration_id

    def modify(self, name, data):
        self.handlers_cache.delete(name)
        integration_record = db.session.query(db.Integration).filter_by(
            company_id=ctx.company_id, name=name
        ).first()
        old_data = deepcopy(integration_record.data)
        for k in old_data:
            if k not in data:
                data[k] = old_data[k]

        integration_record.data = data
        db.session.commit()

    def delete(self, name):
        if name in ('files', 'lightwood'):
            raise Exception('Unable to drop: is system database')

        self.handlers_cache.delete(name)

        # check permanent integration
        if name in self.handler_modules:
            handler = self.handler_modules[name]

            if getattr(handler, 'permanent', False) is True:
                raise Exception('Unable to drop: is permanent integration')

        integration_record = db.session.query(db.Integration).filter_by(company_id=ctx.company_id, name=name).first()

        # if this is ml engine
        engine_models = get_model_records(ml_handler_name=name, deleted_at=None)
        active_models = [m.name for m in engine_models if m.deleted_at is None]
        if len(active_models) > 0:
            raise Exception(f'Unable to drop ml engine with active models: {active_models}')

        # check linked predictors
        models = get_model_records()
        for model in models:
            if (
                model.data_integration_ref is not None
                and model.data_integration_ref.get('type') == 'integration'
                and isinstance(model.data_integration_ref.get('id'), int)
                and model.data_integration_ref['id'] == integration_record.id
            ):
                model.data_integration_ref = None

        # unlink deleted models
        for model in engine_models:
            if model.deleted_at is not None:
                model.integration_id = None

        db.session.delete(integration_record)
        db.session.commit()

    def _get_integration_record_data(self, integration_record, sensitive_info=True):
        if (
            integration_record is None
            or integration_record.data is None
            or isinstance(integration_record.data, dict) is False
        ):
            return None
        data = deepcopy(integration_record.data)
        if data.get('password', None) is None:
            data['password'] = ''

        bundle_path = data.get('secure_connect_bundle')
        mysql_ssl_ca = data.get('ssl_ca')
        mysql_ssl_cert = data.get('ssl_cert')
        mysql_ssl_key = data.get('ssl_key')
        if (
            data.get('type') in ('mysql', 'mariadb')
            and (
                self._is_not_empty_str(mysql_ssl_ca)
                or self._is_not_empty_str(mysql_ssl_cert)
                or self._is_not_empty_str(mysql_ssl_key)
            )
            or data.get('type') in ('cassandra', 'scylla')
            and bundle_path is not None
        ):
            fs_store = FsStore()
            integrations_dir = Config()['paths']['integrations']
            folder_name = f'integration_files_{integration_record.company_id}_{integration_record.id}'
            fs_store.get(
                folder_name,
                base_dir=integrations_dir
            )

        if not sensitive_info:
            if 'password' in data:
                data['password'] = None
            if (
                data.get('type') == 'redis'
                and isinstance(data.get('connection'), dict)
                and 'password' in data['connection']
            ):
                data['connection'] = None

        integration_type = None
        integration_module = self.handler_modules.get(integration_record.engine)
        if hasattr(integration_module, 'type'):
            integration_type = integration_module.type

        class_type = None
        if integration_module is not None and inspect.isclass(integration_module.Handler):
            if issubclass(integration_module.Handler, DatabaseHandler):
                class_type = 'sql'
            if issubclass(integration_module.Handler, APIHandler):
                class_type = 'api'
            if issubclass(integration_module.Handler, BaseMLEngine):
                class_type = 'ml'

        return {
            'id': integration_record.id,
            'name': integration_record.name,
            'type': integration_type,
            'class_type': class_type,
            'engine': integration_record.engine,
            'date_last_update': deepcopy(integration_record.updated_at),
            'connection_data': data
        }

    def get_by_id(self, integration_id, sensitive_info=True):
        integration_record = db.session.query(db.Integration).filter_by(company_id=ctx.company_id, id=integration_id).first()
        return self._get_integration_record_data(integration_record, sensitive_info)

    def get(self, name, sensitive_info=True, case_sensitive=False):
        if case_sensitive:
            integration_record = db.session.query(db.Integration).filter_by(
                company_id=ctx.company_id, name=name
            ).first()
        else:
            integration_record = db.session.query(db.Integration).filter(
                (db.Integration.company_id == ctx.company_id)
                & (func.lower(db.Integration.name) == func.lower(name))
            ).first()
        return self._get_integration_record_data(integration_record, sensitive_info)

    def get_all(self, sensitive_info=True):
        integration_records = db.session.query(db.Integration).filter_by(company_id=ctx.company_id).all()
        integration_dict = {}
        for record in integration_records:
            if record is None or record.data is None:
                continue
            integration_dict[record.name] = self._get_integration_record_data(record, sensitive_info)
        return integration_dict

    def check_connections(self):
        connections = {}
        for integration_name, integration_meta in self.get_all().items():
            handler = self.create_tmp_handler(
                handler_type=integration_meta['engine'],
                connection_data=integration_meta['connection_data']
            )
            status = handler.check_connection()
            connections[integration_name] = status.get('success', False)
        return connections

    def _make_handler_args(self, name: str, handler_type: str, connection_data: dict, integration_id: int = None,
                           file_storage: FileStorage = None, handler_storage: HandlerStorage = None):
        handler_ars = dict(
            name=name,
            integration_id=integration_id,
            connection_data=connection_data,
            file_storage=file_storage,
            handler_storage=handler_storage
        )

        if handler_type == 'files':
            handler_ars['file_controller'] = FileController()
        elif self.handler_modules.get(handler_type, False).type == HANDLER_TYPE.ML:
            handler_ars['handler_controller'] = self
            handler_ars['company_id'] = ctx.company_id

        return handler_ars

    def create_tmp_handler(self, handler_type: str, connection_data: dict) -> object:
        """ Returns temporary handler. That handler does not exist in database.

            Args:
                handler_type (str)
                connection_data (dict)

            Returns:
                Handler object
        """
        handler_meta = self.handlers_import_status[handler_type]
        if not handler_meta["import"]["success"]:
            logger.info(f"to use {handler_type} please install 'pip install mindsdb[{handler_type}]'")

        logger.debug("%s.create_tmp_handler: connection args - %s", self.__class__.__name__, connection_data)
        integration_id = int(time() * 10000)
        file_storage = FileStorage(
            resource_group=RESOURCE_GROUP.INTEGRATION,
            resource_id=integration_id,
            root_dir='tmp',
            sync=False
        )
        handler_storage = HandlerStorage(integration_id, root_dir='tmp', is_temporal=True)
        handler_ars = self._make_handler_args(
            name='tmp_handler',
            handler_type=handler_type,
            connection_data=connection_data,
            integration_id=integration_id,
            file_storage=file_storage,
            handler_storage=handler_storage,
        )

        logger.debug("%s.create_tmp_handler: create a client to db of %s type", self.__class__.__name__, handler_type)
        return DBClient(handler_type, self.handler_modules[handler_type].Handler, **handler_ars)

    def copy_integration_storage(self, integration_id_from, integration_id_to):
        storage_from = HandlerStorage(integration_id_from)
        root_path = ''

        if storage_from.is_empty():
            return None
        folder_from = storage_from.folder_get(root_path)

        storage_to = HandlerStorage(integration_id_to)
        folder_to = storage_to.folder_get(root_path)

        shutil.copytree(folder_from, folder_to, dirs_exist_ok=True)
        storage_to.folder_sync(root_path)

    @profiler.profile()
    def get_handler(self, name, case_sensitive=False):
        handler = self.handlers_cache.get(name)
        if handler is not None:
            return handler

        if case_sensitive:
            integration_record = db.session.query(db.Integration).filter_by(company_id=ctx.company_id, name=name).first()
        else:
            integration_record = db.session.query(db.Integration).filter(
                (db.Integration.company_id == ctx.company_id)
                & (func.lower(db.Integration.name) == func.lower(name))
            ).first()

        integration_data = self._get_integration_record_data(integration_record, True)
        if integration_data is None:
            raise Exception(f"Can't find integration_record for handler '{name}'")
        connection_data = integration_data.get('connection_data', {})
        integration_engine = integration_data['engine']
        integration_name = integration_data['name']
        logger.debug("%s.get_handler: connection_data=%s, engine=%s", self.__class__.__name__, connection_data, integration_engine)

        if integration_engine not in self.handler_modules:
            raise Exception(f"Can't find handler for '{integration_name}' ({integration_engine})")

        integration_meta = self.handlers_import_status[integration_engine]
        if integration_meta["import"]["success"] is False:
            msg = dedent(f'''\
                Handler '{integration_engine}' cannot be used. Reason is:
                    {integration_meta['import']['error_message']}
            ''')
            is_cloud = Config().get('cloud', False)
            if is_cloud is False:
                msg += dedent(f'''

                If error is related to missing dependencies, then try to run command in shell and restart mindsdb:
                    pip install mindsdb[{integration_engine}]
                ''')
            logger.debug(msg)
            raise Exception(msg)

        connection_args = integration_meta.get('connection_args')
        logger.debug("%s.get_handler: connection args - %s", self.__class__.__name__, connection_args)

        file_storage = FileStorage(
            resource_group=RESOURCE_GROUP.INTEGRATION,
            resource_id=integration_record.id,
            sync=True,
        )
        handler_storage = HandlerStorage(integration_record.id)

        if isinstance(connection_args, (dict, OrderedDict)):
            files_to_get = {
                arg_name: arg_value for arg_name, arg_value in connection_data.items()
                if arg_name in connection_args and connection_args.get(arg_name)['type'] == ARG_TYPE.PATH
            }
            if len(files_to_get) > 0:

                for file_name, file_path in files_to_get.items():
                    connection_data[file_name] = file_storage.get_path(file_path)

        handler_ars = self._make_handler_args(
            name=name,
            handler_type=integration_engine,
            connection_data=connection_data,
            integration_id=integration_data['id'],
            file_storage=file_storage,
            handler_storage=handler_storage
        )

        HandlerClass = self.handler_modules[integration_engine].Handler

        if integration_meta.get('type') == HANDLER_TYPE.ML:
            ml_handler_args = {
                'name': handler_ars['name'],
                'integration_id': handler_ars['integration_id'],
                'integration_engine': integration_engine,
                'handler_class': HandlerClass
            }
            logger.info("%s.get_handler: create a ML client, params - %s", self.__class__.__name__, ml_handler_args)
            handler = BaseMLEngineExec(**ml_handler_args)
        else:
            logger.info("%s.get_handler: create a client to db service of %s type, args - %s", self.__class__.__name__, integration_engine, handler_ars)
            handler = HandlerClass(**handler_ars)
            # handler = DBClient(integration_engine, HandlerClass, **handler_ars)
            self.handlers_cache.set(handler)

        return handler

    def reload_handler_module(self, handler_name):
        importlib.reload(self.handler_modules[handler_name])
        try:
            handler_meta = self._get_handler_meta(self.handler_modules[handler_name])
        except Exception as e:
            handler_meta = {
                'import': {
                    'success': False,
                    'error_message': str(e),
                    'dependencies': []
                },
                'name': handler_name
            }

        self.handlers_import_status[handler_meta['name']] = handler_meta

    def _read_dependencies(self, path):
        dependencies = []
        requirements_txt = Path(path).joinpath('requirements.txt')
        if requirements_txt.is_file():
            with open(str(requirements_txt), 'rt') as f:
                dependencies = [x.strip(' \t\n') for x in f.readlines()]
                dependencies = [x for x in dependencies if len(x) > 0]
        return dependencies

    def _get_handler_meta(self, module):
        handler_dir = Path(module.__path__[0])
        handler_folder_name = handler_dir.name
        dependencies = self._read_dependencies(handler_dir)

        self.handler_modules[module.name] = module
        import_error = getattr(module, 'import_error', None)
        handler_meta = {
            'import': {
                'success': import_error is None,
                'folder': handler_folder_name,
                'dependencies': dependencies
            },
            'version': module.version
        }
        if import_error is not None:
            handler_meta['import']['error_message'] = str(import_error)

        # for ml engines, patch the connection_args from the argument probing
        if hasattr(module, 'Handler'):
            handler_class = module.Handler
            try:
                prediction_args = handler_class.prediction_args()
                creation_args = handler_class.creation_args()
                connection_args = {
                    "prediction": prediction_args,
                    "creation_args": creation_args
                }
                setattr(module, 'connection_args', connection_args)
                logger.debug("Patched connection_args for %s", handler_folder_name)
            except Exception as e:
                # do nothing
                logger.debug("Failed to patch connection_args for %s, reason: %s", handler_folder_name, str(e))
        module_attrs = [attr for attr in [
            'connection_args_example',
            'connection_args',
            'description',
            'name',
            'type',
            'title'
        ] if hasattr(module, attr)]

        for attr in module_attrs:
            handler_meta[attr] = getattr(module, attr)

        # region icon
        if hasattr(module, 'icon_path'):
            icon_path = handler_dir.joinpath(module.icon_path)
            handler_meta['icon'] = {
                'name': icon_path.name,
                'type': icon_path.name[icon_path.name.rfind('.') + 1:].lower()
            }
            if handler_meta['icon']['type'] == 'svg':
                with open(str(icon_path), 'rt') as f:
                    handler_meta['icon']['data'] = f.read()
            else:
                with open(str(icon_path), 'rb') as f:
                    handler_meta['icon']['data'] = base64.b64encode(f.read()).decode('utf-8')
        # endregion
        if hasattr(module, 'permanent'):
            handler_meta['permanent'] = module.permanent
        else:
            if handler_meta.get('name') in ('files', 'views', 'lightwood'):
                handler_meta['permanent'] = True
            else:
                handler_meta['permanent'] = False

        return handler_meta

    def _load_handler_modules(self):
        mindsdb_path = Path(importlib.util.find_spec('mindsdb').origin).parent
        handlers_path = mindsdb_path.joinpath('integrations/handlers')

        # edge case: running from tests directory, find_spec finds the base folder instead of actual package
        if not os.path.isdir(handlers_path):
            mindsdb_path = Path(importlib.util.find_spec('mindsdb').origin).parent.joinpath('mindsdb')
            handlers_path = mindsdb_path.joinpath('integrations/handlers')

        self.handler_modules = {}
        self.handlers_import_status = {}
        for handler_dir in handlers_path.iterdir():
            if handler_dir.is_dir() is False or handler_dir.name.startswith('__'):
                continue
            handler_folder_name = str(handler_dir.name)

            try:
                handler_module = importlib.import_module(f'mindsdb.integrations.handlers.{handler_folder_name}')
                handler_meta = self._get_handler_meta(handler_module)
            except Exception as e:
                handler_name = handler_folder_name
                if handler_name.endswith('_handler'):
                    handler_name = handler_name[:-8]
                dependencies = self._read_dependencies(handler_dir)
                handler_meta = {
                    'import': {
                        'success': False,
                        'error_message': str(e),
                        'folder': handler_folder_name,
                        'dependencies': dependencies
                    },
                    'name': handler_name
                }

            self.handlers_import_status[handler_meta['name']] = handler_meta

    def get_handlers_import_status(self):
        return self.handlers_import_status


integration_controller = IntegrationController()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from ..scylla_handler import Handler as ScyllaHandler


class CassandraHandler(ScyllaHandler):
    """
    This handler handles connection and execution of the Cassandra statements.
    """
    name = 'cassandra'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'User name',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'Password',
        'required': True,
        'label': 'Password'
    },
    protocol_version={
        'type': ARG_TYPE.INT,
        'description': 'is not required and defaults to 4.',
        'required': False,
        'label': 'Protocol version'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': ' is the host name or IP address of the Cassandra database.',
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Server port',
        'required': True,
        'label': 'Port'
    },
    keyspace={
        'type': ARG_TYPE.STR,
        'description': ' is the keyspace to connect, the top level container for tables.',
        'required': True,
        'label': 'Keyspace'
    },
    secure_connect_bundle={
        'type': ARG_TYPE.STR,
        'description': 'Path or URL to the secure connect bundle',
        'required': False,
        'label': 'Host'
    }
)



import copy
from typing import List

import mindsdb_sql.planner.utils as utils
from mindsdb_sql.parser.ast import (
    ASTNode,
    BinaryOperation,
    Constant,
    Data,
    Delete,
    Identifier,
    Insert,
    Join,
    Select,
    TableColumn,
    Update,
)

import mindsdb.interfaces.storage.db as db
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import SQLQuery
from mindsdb.api.mysql.mysql_proxy.executor.data_types import ANSWER_TYPE, ExecuteAnswer
from mindsdb.integrations.libs.vectordatabase_handler import TableField


class KnowledgeBaseController:
    """
    Knowledge bae controller handles all
    db related operations for knowledge bases
    """

    def __init__(self, session) -> None:
        self.executor = KnowledgeBaseExecutor(session=session)
        self.session = session

    def is_knowledge_base(self, identifier: Identifier) -> bool:
        """
        Decide if the identifier is a knowledge base
        """
        return self.executor.is_knowledge_base(identifier)

    def execute_query(self, query: ASTNode) -> ExecuteAnswer:
        """
        Execute a parsed query and return the result
        """
        if isinstance(query, Select):
            return self.executor.select_from_kb(query)
        elif isinstance(query, Insert):
            return self.executor.insert_into_kb(query)
        elif isinstance(query, Delete):
            return self.executor.delete_from_kb(query)
        elif isinstance(query, Update):
            return self.executor.update_kb(query)
        else:
            raise NotImplementedError()

    def add(
        self,
        name: str,
        project_id: str,
        vector_database_id: str,
        vector_database_table_name: str,
        embedding_model_id: str,
        params: dict,
        if_not_exists: bool = False,
    ) -> int:
        """
        Add a new knowledge base to the database
        """
        # check if knowledge base already exists
        try:
            kb = self.get(name, project_id)
        except ValueError:
            # knowledge base does not exist
            kb = db.KnowledgeBase(
                name=name,
                project_id=project_id,
                vector_database_id=vector_database_id,
                vector_database_table=vector_database_table_name,
                embedding_model_id=embedding_model_id,
                params=params,
            )
            db.session.add(kb)
            db.session.commit()
            return kb.id

        # kb already exists
        if if_not_exists:
            return kb.id
        else:
            raise Exception(f"Knowledge base already exists: {name}")

    def delete(self, name: str, project_id: str, if_exists: bool = False) -> None:
        """
        Delete a knowledge base from the database
        """
        # check if knowledge base exists
        try:
            kb = self.get(name, project_id)
        except ValueError:
            # knowledge base does not exist
            if if_exists:
                return
            else:
                raise Exception(f"Knowledge base does not exist: {name}")

        # kb exists
        db.session.delete(kb)
        db.session.commit()

    def get(self, name: str, project_id: str) -> db.KnowledgeBase:
        """
        Get a knowledge base from the database
        by name + project_id
        """
        kb = (
            db.session.query(db.KnowledgeBase)
            .filter_by(
                name=name,
                project_id=project_id,
            )
            .first()
        )
        if kb is None:
            raise ValueError(f"Knowledge base not found: {name}")
        return kb

    def get_by_id(self, id: str) -> db.KnowledgeBase:
        """
        Get a knowledge base from the database
        by id
        """
        kb = (
            db.session.query(db.KnowledgeBase)
            .filter_by(
                id=id,
            )
            .first()
        )
        if kb is None:
            raise ValueError(f"Knowledge base not found: {id}")
        return kb

    def list(self, project_id: str) -> List[db.KnowledgeBase]:
        """
        List all knowledge bases from the database
        belonging to a project
        """
        kbs = (
            db.session.query(db.KnowledgeBase)
            .filter_by(
                project_id=project_id,
            )
            .all()
        )
        return kbs

    def update(self, name: str, project_id: str, **kwargs) -> db.KnowledgeBase:
        """
        Update a knowledge base from the database
        """
        raise NotImplementedError()

    def update_by_id(self, id: str, **kwargs) -> db.KnowledgeBase:
        """
        Update a knowledge base from the database
        """
        raise NotImplementedError()

    def list_kb_context_entry(self) -> List[dict]:
        """
        List all knowledge base context entries
        """
        # TODO: this is n+1 query, need to optimize
        kbs = db.session.query(
            db.KnowledgeBase,
        ).all()
        kb_context_entries = [self.get_kb_context_entry_by_id(kb.id) for kb in kbs]
        return kb_context_entries

    def get_kb_context_entry_by_id(self, id: str) -> dict:
        """
        Get the knowledge base context entry
        in the format of
        {
            "name": "my_kb",
            "type": "knowledge_base",
            "model": "mindsdb.my_model",
            "storage": "my_chromadb.my_table",
            "search_vector_field": "search_vector",
            "embeddings_field": "embeddings",
            "content_field": "content",
        }
        """
        kb = self.get_by_id(id)
        name = kb.name
        type_ = "knowledge_base"
        search_vector_field = "search_vector"

        # get model
        model_id = kb.embedding_model_id
        # get the input columns
        model = (
            db.session.query(db.Predictor)
            .filter_by(
                id=model_id,
            )
            .first()
        )

        assert model is not None, f"Model not found: {model_id}"
        model_project = (
            db.session.query(db.Project)
            .filter_by(
                id=model.project_id,
            )
            .first()
        )

        assert model_project is not None, f"Project not found: {model.project_id}"
        model_name = f"{model_project.name}.{model.name}"

        # get the output columns
        # describe the model
        args_df = self.session.model_controller.describe_model(
            session=self.session,
            project_name=model_project.name,
            model_name=model.name,
            attribute="args",
        )
        args_df.set_index("key", inplace=True)
        embeddings_field = args_df.loc["target", "value"]
        # TODO: assuming there is only one input column
        # need to support multiple input columns
        content_field = args_df.loc["input_columns", "value"][0]

        # get the vector database name + table
        database_id = kb.vector_database_id
        table_name = kb.vector_database_table
        # get the database object
        database = (
            db.session.query(db.Integration)
            .filter_by(
                id=database_id,
            )
            .first()
        )

        assert database is not None, f"Database not found: {database_id}"

        storage = f"{database.name}.{table_name}"
        return {
            "name": name,
            "type": type_,
            "model": model_name,
            "storage": storage,
            "search_vector_field": search_vector_field,
            "embeddings_field": embeddings_field,
            "content_field": content_field,
        }


class KnowledgeBaseExecutor:
    """
    Knowledge base executor handles all
    sql related operations for knowledge bases
    """

    MODEL_FIELD = "model"
    STORAGE_FIELD = "storage"
    SEARCH_QUERY = "search_query"

    def __init__(self, session) -> None:
        self.session = session

    def _get_knowledge_base_metadata(self, identifier: Identifier) -> dict:
        """
        Get the metadata of a knowledge base
        """
        name_parts = list(identifier.parts)
        name = name_parts[-1]
        if len(name_parts) > 1:
            namespace = name_parts[-2]
        else:
            namespace = self.session.database

        # query the db
        project_id = self.session.database_controller.get_project(namespace).id

        kb = self.session.kb_controller.get(
            name=name,
            project_id=project_id,
        )

        kb_metadata = self.session.kb_controller.get_kb_context_entry_by_id(
            id=kb.id,
        )

        return kb_metadata

    def is_knowledge_base(self, identifier: Identifier) -> bool:
        """
        Check if the identifier is a knowledge base
        """
        try:
            self._get_knowledge_base_metadata(identifier)
            return True
        except (ValueError, AttributeError):
            return False

    def select_from_kb(self, query: Select) -> ExecuteAnswer:
        """
        Handle the select query
        We do the following translation logics:
        1. Select from the underlying storage table
        2. If a search query clause is provided in where, we
            substitute the search query clause with a nested select
            from the underlying model query
        """
        knowledge_base_metadata = self._get_knowledge_base_metadata(query.from_table)
        vector_database_table = knowledge_base_metadata[self.STORAGE_FIELD]
        model_name = knowledge_base_metadata[self.MODEL_FIELD]

        CONTENT_FIELD = (
            knowledge_base_metadata.get("content_field") or TableField.CONTENT.value
        )
        EMBEDDINGS_FIELD = (
            knowledge_base_metadata.get("embeddings_field")
            or TableField.EMBEDDINGS.value
        )
        SEARCH_VECTOR_FIELD = (
            knowledge_base_metadata.get("search_vector_field")
            or TableField.SEARCH_VECTOR.value
        )

        is_search_query_present = False

        def find_search_query(node, **kwargs):
            nonlocal is_search_query_present
            if isinstance(node, Identifier) and node.parts[-1] == self.SEARCH_QUERY:
                is_search_query_present = True

        # decide predictor is needed in the query
        # by detecting if a where clause involving field SEARCH_QUERY is present
        # if yes, then we need to add additional step to the plan
        # to apply the predictor to the search query
        utils.query_traversal(query.where, callback=find_search_query)

        if not is_search_query_present:
            # dispatch to the underlying storage table
            query.from_table = Identifier(vector_database_table)
        else:
            # rewrite the where clause
            # search_query = 'some text'
            # ->
            # search_vector = (select embeddings from model_name where content = 'some text')
            def rewrite_search_query_clause(node, **kwargs):
                if isinstance(node, BinaryOperation):
                    if node.args[0] == Identifier(self.SEARCH_QUERY):
                        node.args[0] = Identifier(SEARCH_VECTOR_FIELD)
                        node.args[1] = Select(
                            targets=[Identifier(EMBEDDINGS_FIELD)],
                            from_table=Identifier(model_name),
                            where=BinaryOperation(
                                op="=", args=[Identifier(CONTENT_FIELD), node.args[1]]
                            ),
                        )

            utils.query_traversal(query.where, callback=rewrite_search_query_clause)

            # dispatch to the underlying storage table
            query.from_table = Identifier(vector_database_table)
        sql_query = SQLQuery(sql=query, session=self.session, execute=True)
        data = sql_query.fetch()

        return ExecuteAnswer(
            answer_type=ANSWER_TYPE.TABLE,
            columns=sql_query.columns_list,
            data=data["result"],
        )

    def insert_into_kb(self, query: Insert):
        """
        Handle the insert query
        We do the following translation logics:
        1. Insert into the underlying storage table
        2. If a select query is present, we join the select query
            with a model to get the embeddings column
        3. If values are present, we wrap the values in ast.Data
            join them with the model to get the embeddings column
        """
        metadata = self._get_knowledge_base_metadata(query.table)
        EMBEDDINGS_FIELD = (
            metadata.get("embeddings_field") or TableField.EMBEDDINGS.value
        )

        vector_database_table = metadata[self.STORAGE_FIELD]
        model_name = metadata[self.MODEL_FIELD]

        query.table = Identifier(vector_database_table)

        if query.from_select is not None:
            # detect if embeddings field is present in the columns list
            # if so, we do not need to apply the predictor
            # if not, we need to join the select with the model table
            is_embeddings_field_present = False

            def find_embeddings_field(node, **kwargs):
                nonlocal is_embeddings_field_present
                if isinstance(node, Identifier) and node.parts[-1] == EMBEDDINGS_FIELD:
                    is_embeddings_field_present = True

            utils.query_traversal(query.columns, callback=find_embeddings_field)

            if is_embeddings_field_present:
                return self.plan_insert(query)

            # rewrite the select statement
            # to join with the model table

            select: Select = query.from_select
            select.targets.append(Identifier(EMBEDDINGS_FIELD))
            select.from_table = Select(
                targets=copy.deepcopy(select.targets),
                from_table=Join(
                    left=select.from_table,
                    right=Identifier(model_name),
                    join_type="JOIN",
                ),
            )

            # append the embeddings field to the columns list
            if query.columns:
                query.columns.append(Identifier(EMBEDDINGS_FIELD))

            return self.plan_insert(query)
        else:
            if not query.columns:
                raise Exception(
                    "Insert into knowledge base requires a select query or a list of columns"
                )

            keys = [column.name for column in query.columns]
            is_embeddings_field_present = EMBEDDINGS_FIELD in keys

            query.table = Identifier(vector_database_table)
            # directly dispatch to the underlying storage table
            if is_embeddings_field_present:
                return self.plan_insert(query)

            # if the embeddings field is not present in the columns list
            # we need to wrap values in ast.Data
            # join it with a model table
            # modify the query using from_table
            # and dispatch to the underlying storage table

            records = []
            _unwrap_constant_or_self = (
                lambda node: node.value if isinstance(node, Constant) else node
            )
            for row in query.values:
                records.append(dict(zip(keys, map(_unwrap_constant_or_self, row))))

            data = Data(records, alias=Identifier("data"))
            predictor_select = Select(
                targets=[Identifier(col.name) for col in query.columns]
                + [Identifier(EMBEDDINGS_FIELD)],
                from_table=Join(
                    left=data, right=Identifier(model_name), join_type="JOIN"
                ),
            )

            query.columns += [TableColumn(name=EMBEDDINGS_FIELD)]
            query.from_select = predictor_select
            query.values = None

            _ = SQLQuery(sql=query, session=self.session, execute=True)
            return ExecuteAnswer(
                answer_type=ANSWER_TYPE.OK,
            )

    def plan_insert(self, query: Insert):
        # TODO: to be implemented
        raise NotImplementedError()

    def delete_from_kb(self, query: Delete):
        metadata = self._get_knowledge_base_metadata(query.table)

        vector_database_table = metadata[self.STORAGE_FIELD]
        query.table = Identifier(vector_database_table)

        _ = SQLQuery(sql=query, session=self.session, execute=True)
        return ExecuteAnswer(
            answer_type=ANSWER_TYPE.OK,
        )

    def update_kb(self, query: Update) -> ExecuteAnswer:
        # TODO: to be implemented
        raise NotImplementedError()



from typing import Any
import ast as py_ast

import pandas as pd
from mindsdb_sql.parser.ast import ASTNode, Select, Insert, Update, Delete
from mindsdb_sql.parser.ast.select.identifier import Identifier

from mindsdb.integrations.libs.base import BaseHandler
from mindsdb.integrations.libs.api_handler_exceptions import TableAlreadyExists, TableNotFound

from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE
)


class FuncParser:

    def from_string(self, query_string):

        body = py_ast.parse(query_string.strip(), mode='eval').body

        if not isinstance(body, py_ast.Call):
            raise RuntimeError(f'Api function not found {query_string}')

        fnc_name = body.func.id

        params = {}
        for keyword in body.keywords:
            name = keyword.arg
            value = self.process(keyword.value)

            params[name] = value

        return fnc_name, params

    def process(self, node):

        if isinstance(node, py_ast.List):
            elements = []
            for node2 in node.elts:
                elements.append(self.process(node2))
            return elements

        if isinstance(node, py_ast.Dict):

            keys = []
            for node2 in node.keys:
                if isinstance(node2, py_ast.Constant):
                    value = node2.value
                elif isinstance(node2, py_ast.Str):  # py37
                    value = node2.s
                else:
                    raise NotImplementedError(f'Unknown dict key {node2}')

                keys.append(value)

            values = []
            for node2 in node.values:
                values.append(self.process(node2))

            return dict(zip(keys, values))

        if isinstance(node, py_ast.Name):
            # special attributes
            name = node.id
            if name == 'true':
                return True
            elif name == 'false':
                return False
            elif name == 'null':
                return None

        if isinstance(node, py_ast.Constant):
            return node.value

        # ---- python 3.7 objects -----
        if isinstance(node, py_ast.Str):
            return node.s

        if isinstance(node, py_ast.Num):
            return node.n

        # -----------------------------

        if isinstance(node, py_ast.UnaryOp):
            if isinstance(node.op, py_ast.USub):
                value = self.process(node.operand)
                return -value

        raise NotImplementedError(f'Unknown node {node}')


class APITable():

    def __init__(self, handler):
        self.handler = handler

    def select(self, query: ASTNode) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Select

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def insert(self, query: ASTNode) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """
        raise NotImplementedError()

    def update(self, query: ASTNode) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update
        Returns:
            None
        """
        raise NotImplementedError()

    def delete(self, query: ASTNode) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Delete

        Returns:
            None
        """
        raise NotImplementedError()

    def get_columns(self) -> list:
        """Maps the columns names from the API call resource

        Returns:
            List
        """
        raise NotImplementedError()


class APIHandler(BaseHandler):
    """
    Base class for handlers associated to the applications APIs (e.g. twitter, slack, discord  etc.)
    """

    def __init__(self, name: str):
        super().__init__(name)
        """ constructor
        Args:
            name (str): the handler name
        """

        self._tables = {}

    def _register_table(self, table_name: str, table_class: Any):
        """
        Register the data resource. For e.g if you are using Twitter API it registers the `tweets` resource from `/api/v2/tweets`.
        """
        if table_name in self._tables:
            raise TableAlreadyExists(f"Table with name {table_name} already exists for this handler")
        self._tables[table_name] = table_class

    def _get_table(self, name: Identifier):
        """
        Check if the table name was added to the the _register_table
        Args:
            name (Identifier): the table name
        """
        name = name.parts[-1]
        if name not in self._tables:
            raise TableNotFound(f'Table not found: {name}')
        return self._tables[name]

    def query(self, query: ASTNode):

        if isinstance(query, Select):
            result = self._get_table(query.from_table).select(query)
        elif isinstance(query, Update):
            result = self._get_table(query.table).update(query)
        elif isinstance(query, Insert):
            result = self._get_table(query.table).insert(query)
        elif isinstance(query, Delete):
            result = self._get_table(query.table).delete(query)
        else:
            raise NotImplementedError

        if result is None:
            return Response(RESPONSE_TYPE.OK)
        elif isinstance(result, pd.DataFrame):
            return Response(RESPONSE_TYPE.TABLE, result)
        else:
            raise NotImplementedError

    def get_columns(self, table_name: str) -> Response:
        """
        Returns a list of entity columns
        Args:
            table_name (str): the table name
        Returns:
            RESPONSE_TYPE.TABLE
        """

        result = self._get_table(Identifier(table_name)).get_columns()

        df = pd.DataFrame(result, columns=['Field'])
        df['Type'] = 'str'

        return Response(RESPONSE_TYPE.TABLE, df)

    def get_tables(self) -> Response:
        """
        Return list of entities
        Returns:
            RESPONSE_TYPE.TABLE
        """
        result = list(self._tables.keys())

        df = pd.DataFrame(result, columns=['table_name'])
        df['table_type'] = 'BASE TABLE'

        return Response(RESPONSE_TYPE.TABLE, df)


class APIChatHandler(APIHandler):

    def get_chat_config(self):
        """Return configuration to connect to chatbot

        Returns:
            Dict
        """
        raise NotImplementedError()

    def get_my_user_name(self) -> list:
        """Return configuration to connect to chatbot

        Returns:
            Dict
        """
        raise NotImplementedError()



"""jobs

Revision ID: b6d0a47294ac
Revises: 459218b0844c
Create Date: 2023-02-02 11:33:06.283055

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa

# revision identifiers, used by Alembic.
revision = 'b6d0a47294ac'
down_revision = '459218b0844c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'jobs_history',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('job_id', sa.Integer(), nullable=True),
        sa.Column('start_at', sa.DateTime(), nullable=True),
        sa.Column('end_at', sa.DateTime(), nullable=True),
        sa.Column('error', sa.String(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('job_id', 'start_at', name='uniq_job_history_job_id_start')
    )
    op.create_table(
        'jobs',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('query_str', sa.String(), nullable=False),
        sa.Column('start_at', sa.DateTime(), nullable=True),
        sa.Column('end_at', sa.DateTime(), nullable=True),
        sa.Column('next_run_at', sa.DateTime(), nullable=True),
        sa.Column('schedule_str', sa.String(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('jobs')
    op.drop_table('jobs_history')
    # ### end Alembic commands ###



from typing import Optional

import dill
import pandas as pd

from mindsdb.integrations.handlers.sentence_transformer_handler.settings import (
    Parameters,
    df_to_documents,
    load_embeddings_model,
)
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

logger = log.get_log(__name__)


class SentenceTransformerHandler(BaseMLEngine):
    name = "sentence transformer"

    def create(self, target, df=None, args=None, **kwargs):
        """creates embeddings model and persists"""

        args = args["using"]

        valid_args = Parameters(**args)

        model = load_embeddings_model(valid_args.embeddings_model_name)

        self.model_storage.file_set("model", dill.dumps(model))
        self.model_storage.json_set("args", valid_args.dict())

    def predict(self, df, args=None):
        """loads persisted embeddings model and gets embeddings on input text column(s)"""

        args = args["predict_params"]
        columns = args.get("columns")

        if columns:
            if isinstance(args["columns"], str):
                columns = [columns]

        else:
            logger.info("no columns specified, all columns from input will be embedded")

            columns = df.columns

        documents = df_to_documents(df=df, page_content_columns=columns)

        model = dill.loads(self.model_storage.file_get("model"))

        embeddings = []

        for _, document in enumerate(documents):
            _embeddings = model.encode(document.text).tolist()
            embeddings.append(_embeddings)

        embeddings_df = pd.DataFrame(data={"embeddings": embeddings})

        return embeddings_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:

        args = self.model_storage.json_get("args")

        if attribute == "args":
            return pd.DataFrame(args.items(), columns=["key", "value"])



import sys


def format_exception_error(exception):
    try:
        exception_type, _exception_object, exception_traceback = sys.exc_info()
        filename = exception_traceback.tb_frame.f_code.co_filename
        line_number = exception_traceback.tb_lineno
        error_message = f'{exception_type.__name__}: {exception}, raised at: {filename}#{line_number}'
    except Exception:
        error_message = str(exception)
    return error_message



import unittest
from mindsdb.integrations.handlers.sqreamdb_handler.sqreamdb_handler import SQreamDBHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class SQreamDBHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "127.0.0.1",
                "port": "5000",
                "user": "sqream",
                "password": "sqream",
                "database": "master"
            }
        }
        cls.handler = SQreamDBHandler('test_sqreamdb_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_drop_table(self):
        res = self.handler.query("DROP TABLE IF EXISTS LOVE")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_2_create_table(self):
        res = self.handler.query("CREATE TABLE IF NOT EXISTS LOVE (LOVER varchar(20))")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_3_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR
 
    def test_4_select_query(self):
        query = "SELECT * FROM AUTHORS"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_check_connection(self):
         self.handler.check_connection()

        
if __name__ == '__main__':
    unittest.main()


from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'connectionStatus': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        res = {
            'authInfo': {
                'authenticatedUsers': [],
                'authenticatedUserRoles': []
            },
            'ok': 1
        }
        if helpers.is_true(query.get('showPrivileges')):
            res['authInfo']['authenticatedUserPrivileges'] = []

        return res


responder = Responce()



import ast
import difflib
import uuid
from enum import Enum
from typing import Any, List, Optional

import pandas as pd
from mindsdb_sql.parser.ast import (
    BinaryOperation,
    Constant,
    CreateTable,
    Delete,
    DropTables,
    Insert,
    Select,
    Star,
    Tuple,
    Update,
)
from mindsdb_sql.parser.ast.base import ASTNode
from pydantic import BaseModel, Extra, root_validator

from mindsdb.integrations.libs.response import RESPONSE_TYPE, HandlerResponse
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.utilities.log import get_log

from ..utilities.sql_utils import query_traversal
from .base import BaseHandler

LOG = get_log("VectorStoreHandler")


class FilterOperator(Enum):
    """
    Enum for filter operators.
    """

    EQUAL = "="
    NOT_EQUAL = "!="
    LESS_THAN = "<"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN = ">"
    GREATER_THAN_OR_EQUAL = ">="
    IN = "IN"
    NOT_IN = "NOT IN"
    BETWEEN = "BETWEEN"
    NOT_BETWEEN = "NOT BETWEEN"
    LIKE = "LIKE"
    NOT_LIKE = "NOT LIKE"
    IS_NULL = "IS NULL"
    IS_NOT_NULL = "IS NOT NULL"


class FilterCondition:
    """
    Base class for filter conditions.
    """

    def __init__(self, column: str, op: FilterOperator, value: Any):
        self.column = column
        self.op = op
        self.value = value

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, FilterCondition):
            return (
                self.column == __value.column
                and self.op == __value.op
                and self.value == __value.value
            )
        else:
            return False

    def __repr__(self) -> str:
        return f"""
            FilterCondition(
                column={self.column},
                op={self.op},
                value={self.value}
            )
        """


class TableField(Enum):
    """
    Enum for table fields.
    """

    ID = "id"
    CONTENT = "content"
    EMBEDDINGS = "embeddings"
    METADATA = "metadata"
    SEARCH_VECTOR = "search_vector"
    DISTANCE = "distance"


class VectorStoreHandlerConfig(BaseModel):
    """
    Configuration for VectorStoreHandler.
    """

    vector_store: str
    persist_directory: str = None
    host: str = None
    port: int = None
    url: str = None

    class Config:
        extra = Extra.forbid

    @root_validator(pre=True, allow_reuse=True)
    def check_param_typos(cls, values):
        """Check if there are any typos in the parameters."""

        expected_params = cls.__fields__.keys()
        for key in values.keys():
            if key not in expected_params:
                close_matches = difflib.get_close_matches(
                    key, expected_params, cutoff=0.4
                )
                if close_matches:
                    raise ValueError(
                        f"Unexpected parameter '{key}'. Did you mean '{close_matches[0]}'?"
                    )
                else:
                    raise ValueError(f"Unexpected parameter '{key}'.")
        return values

    @root_validator(allow_reuse=True)
    def check_config(cls, values):
        """Check if config is valid."""

        vector_store = values.get("vector_store")
        host = values.get("host")
        port = values.get("port")
        url = values.get("url")
        persist_directory = values.get("persist_directory")

        if host and not port:
            raise ValueError(
                f"For {vector_store} handler - if host is provided, port must also be provided."
            )

        if port and not host:
            raise ValueError(
                f"For {vector_store} handler - if port is provided, host must also be provided."
            )

        if port and host and (url or persist_directory):
            raise ValueError(
                f"For {vector_store} handler - if host and port are provided, "
                f"url and persistence_folder should not be provided."
            )

        if url and (host or port or persist_directory):
            raise ValueError(
                f"For {vector_store} handler - if url is provided, host, port, "
                f"persist_directory should not be provided."
            )

        if persist_directory and (url or host or port):
            raise ValueError(
                f"For {vector_store} handler - if persistence_folder is provided, "
                f"url, host, port should not be provided."
            )

        return values


class VectorStoreHandler(BaseHandler):
    """
    Base class for handlers associated to vector databases.
    """

    SCHEMA = [
        {
            "name": TableField.ID.value,
            "data_type": "string",
        },
        {
            "name": TableField.CONTENT.value,
            "data_type": "string",
        },
        {
            "name": TableField.EMBEDDINGS.value,
            "data_type": "list",
        },
        {
            "name": TableField.METADATA.value,
            "data_type": "json",
        },
    ]

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        kwargs["connection_data"].pop("password", None)
        self.handler_storage = HandlerStorage(kwargs.get("integration_id"))

        _config = kwargs.get("connection_data")
        _config["vector_store"] = name

        self.config = VectorStoreHandlerConfig(**_config)

        self.is_connected = False

        if self.config.persist_directory and not self.handler_storage.is_temporal:
            # get full persistence directory from handler storage
            self.persist_directory = self.handler_storage.folder_get(
                self.config.persist_directory
            )

    def __del__(self):
        if self.is_connected is True:
            if self.persist_directory:
                # sync folder to handler storage
                self.handler_storage.folder_sync(self.persist_directory)

            self.disconnect()

    def disconnect(self):
        raise NotImplementedError()

    def _value_or_self(self, value):
        if isinstance(value, Constant):
            return value.value
        else:
            return value

    def _extract_conditions(self, where_statement) -> Optional[List[FilterCondition]]:
        conditions = []
        # parse conditions
        if where_statement is not None:
            # dfs to get all binary operators in the where statement
            def _extract_comparison_conditions(node, **kwargs):
                if isinstance(node, BinaryOperation):
                    # if the op is and, continue
                    # TODO: need to handle the OR case
                    if node.op.upper() == "AND":
                        return
                    op = FilterOperator(node.op.upper())
                    # unquote the left hand side
                    left_hand = node.args[0].parts[-1].strip("`")
                    if isinstance(node.args[1], Constant):
                        if left_hand == TableField.SEARCH_VECTOR.value:
                            right_hand = ast.literal_eval(node.args[1].value)
                        else:
                            right_hand = node.args[1].value
                    elif isinstance(node.args[1], Tuple):
                        # Constant could be actually a list i.e. [1.2, 3.2]
                        right_hand = [
                            ast.literal_eval(item.value)
                            if isinstance(item, Constant)
                            and not isinstance(item.value, list)
                            else item.value
                            for item in node.args[1].items
                        ]
                    else:
                        raise Exception(f"Unsupported right hand side: {node.args[1]}")
                    conditions.append(
                        FilterCondition(column=left_hand, op=op, value=right_hand)
                    )

            query_traversal(where_statement, _extract_comparison_conditions)

            # try to treat conditions that are not in TableField as metadata conditions
            for condition in conditions:
                if not self._is_condition_allowed(condition):
                    condition.column = (
                        TableField.METADATA.value + "." + condition.column
                    )

        else:
            conditions = None

        return conditions

    def _is_columns_allowed(self, columns: List[str]) -> bool:
        """
        Check if columns are allowed.
        """
        allowed_columns = set([col["name"] for col in self.SCHEMA])
        return set(columns).issubset(allowed_columns)

    def _is_condition_allowed(self, condition: FilterCondition) -> bool:
        allowed_field_values = set([field.value for field in TableField])
        if condition.column in allowed_field_values:
            return True
        else:
            # check if column is a metadata column
            if condition.column.startswith(TableField.METADATA.value):
                return True
            else:
                return False

    def _dispatch_create_table(self, query: CreateTable) -> HandlerResponse:
        """
        Dispatch create table query to the appropriate method.
        """
        # parse key arguments
        table_name = query.name.parts[-1]
        if_not_exists = getattr(query, "if_not_exists", False)
        return self.create_table(table_name, if_not_exists=if_not_exists)

    def _dispatch_drop_table(self, query: DropTables) -> HandlerResponse:
        """
        Dispatch drop table query to the appropriate method.
        """
        # parse key arguments
        table_names = [table.parts[-1] for table in query.tables]
        if_exists = getattr(query, "if_exists", False)
        for table_name in table_names:
            self.drop_table(table_name, if_exists=if_exists)
        return HandlerResponse(resp_type=RESPONSE_TYPE.OK)

    def _dispatch_insert(self, query: Insert) -> HandlerResponse:
        """
        Dispatch insert query to the appropriate method.
        """
        # parse key arguments
        table_name = query.table.parts[-1]
        columns = [column.name for column in query.columns]

        if not self._is_columns_allowed(columns):
            raise Exception(
                f"Columns {columns} not allowed."
                f"Allowed columns are {[col['name'] for col in self.SCHEMA]}"
            )

        # get id column if it is present
        if "id" in columns:
            id_col_index = columns.index("id")
            ids = [self._value_or_self(row[id_col_index]) for row in query.values]
        else:
            ids = [uuid.uuid4().hex for _ in query.values]

        # get content column if it is present
        if TableField.CONTENT.value in columns:
            content_col_index = columns.index("content")
            content = [
                self._value_or_self(row[content_col_index]) for row in query.values
            ]
        else:
            content = None

        # get embeddings column if it is present
        if TableField.EMBEDDINGS.value in columns:
            embeddings_col_index = columns.index("embeddings")
            embeddings = [
                ast.literal_eval(self._value_or_self(row[embeddings_col_index]))
                for row in query.values
            ]
        else:
            raise Exception("Embeddings column is required!")

        if TableField.METADATA.value in columns:
            metadata_col_index = columns.index("metadata")
            metadata = [
                ast.literal_eval(self._value_or_self(row[metadata_col_index]))
                for row in query.values
            ]
        else:
            metadata = None

        # create dataframe
        data = pd.DataFrame(
            {
                TableField.ID.value: ids,
                TableField.CONTENT.value: content,
                TableField.EMBEDDINGS.value: embeddings,
                TableField.METADATA.value: metadata,
            }
        )

        # dispatch insert
        return self.insert(table_name, data, columns=columns)

    def _dispatch_update(self, query: Update) -> HandlerResponse:
        """
        Dispatch update query to the appropriate method.
        """
        raise NotImplementedError("Update query is not supported!")

    def _dispatch_delete(self, query: Delete) -> HandlerResponse:
        """
        Dispatch delete query to the appropriate method.
        """
        # parse key arguments
        table_name = query.table.parts[-1]
        where_statement = query.where
        conditions = self._extract_conditions(where_statement)

        # dispatch delete
        return self.delete(table_name, conditions=conditions)

    def _dispatch_select(self, query: Select) -> HandlerResponse:
        """
        Dispatch select query to the appropriate method.
        """
        # parse key arguments
        table_name = query.from_table.parts[-1]
        # if targets are star, select all columns
        if isinstance(query.targets[0], Star):
            columns = [col["name"] for col in self.SCHEMA]
        else:
            columns = [col.parts[-1] for col in query.targets]

        if not self._is_columns_allowed(columns):
            raise Exception(
                f"Columns {columns} not allowed."
                f"Allowed columns are {[col['name'] for col in self.SCHEMA]}"
            )

        # check if columns are allowed
        where_statement = query.where
        conditions = self._extract_conditions(where_statement)

        # get offset and limit
        offset = query.offset.value if query.offset is not None else None
        limit = query.limit.value if query.limit is not None else None

        # dispatch select
        return self.select(
            table_name,
            columns=columns,
            conditions=conditions,
            offset=offset,
            limit=limit,
        )

    def _dispatch(self, query: ASTNode) -> HandlerResponse:
        """
        Parse and Dispatch query to the appropriate method.
        """
        dispatch_router = {
            CreateTable: self._dispatch_create_table,
            DropTables: self._dispatch_drop_table,
            Insert: self._dispatch_insert,
            Update: self._dispatch_update,
            Delete: self._dispatch_delete,
            Select: self._dispatch_select,
        }
        if type(query) in dispatch_router:
            return dispatch_router[type(query)](query)
        else:
            raise NotImplementedError(f"Query type {type(query)} not implemented.")

    def query(self, query: ASTNode) -> HandlerResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc

        Returns:
            HandlerResponse
        """
        return self._dispatch(query)

    def create_table(self, table_name: str, if_not_exists=True) -> HandlerResponse:
        """Create table

        Args:
            table_name (str): table name
            if_not_exists (bool): if True, do nothing if table exists

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def drop_table(self, table_name: str, if_exists=True) -> HandlerResponse:
        """Drop table

        Args:
            table_name (str): table name
            if_exists (bool): if True, do nothing if table does not exist

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def insert(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ) -> HandlerResponse:
        """Insert data into table

        Args:
            table_name (str): table name
            data (pd.DataFrame): data to insert
            columns (List[str]): columns to insert

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def update(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ) -> HandlerResponse:
        """Update data in table

        Args:
            table_name (str): table name
            data (pd.DataFrame): data to update
            columns (List[str]): columns to update

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ) -> HandlerResponse:
        """Delete data from table

        Args:
            table_name (str): table name
            conditions (List[FilterCondition]): conditions to delete

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> HandlerResponse:
        """Select data from table

        Args:
            table_name (str): table name
            columns (List[str]): columns to select
            conditions (List[FilterCondition]): conditions to select

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def get_columns(self, table_name: str) -> HandlerResponse:
        # return a fixed set of columns
        data = pd.DataFrame(self.SCHEMA)
        data.columns = ["COLUMN_NAME", "DATA_TYPE"]
        return HandlerResponse(
            data_frame=data,
        )



POSTGRES_ERROR_CODES = {
    "CLASS_00": {
        "successful_completion": "00000"},
    "CLASS_01": {
        "warning": "01000",
        "dynamic_result_sets_returned": "0100C",
        "implicit_zero_bit_padding": "01008",
        "null_value_eliminated_in_set_function": "01003",
        "privilege_not_granted": "01007",
        "privilege_not_revoked": "01006",
        "string_data_right_truncation": "01004",
        "deprecated_feature": "01P01"},
    "CLASS_02": {
        "no_data": "02000",
        "no_additional_dynamic_result_sets_returned": "02001"
    },
    "CLASS_03": {
        "sql_statement_not_yet_complete": "03000",
    },
    "CLASS_08": {
        "connection_exception": "08000",
        "connection_does_not_exist": "08003",
        "connection_failure": "08006",
        "sqlclient_unable_to_establish_sqlconnection": "08001",
        "sqlserver_rejected_establishment_of_sqlconnection": "08004",
        "transaction_resolution_unknown": "08007",
        "protocol_violation": "08P01"},
    "CLASS_09": {

        "triggered_action_exception": "09000",
    },
    "CLASS_0A": {
        "feature_not_supported": "0A000"},
    "CLASS_0B": {
        "invalid_transaction_initiation": "0B000"},
    "CLASS_0F": {
        "locator_exception": "0F000",
        "invalid_locator_specification": "0F001",
    },
    "CLASS_0L": {
        "invalid_grantor": "0L000",
        "invalid_grant_operation": "0LP01",
    },
    "CLASS_0P": {
        "invalid_role_specification": "0P000",
    },
    "CLASS_0Z": {
        "diagnostics_exception": "0Z000",
        "stacked_diagnostics_accessed_without_active_handler": "0Z002"},
    "CLASS_20": {
        "case_not_found": "20000"},

    "CLASS_21": {
        "cardinality_violation": "21000"},
    "CLASS_22": {
        "data_exception": "22000",
        "array_subscript_error": "2202E",
        "character_not_in_repertoire": "22021",
        "datetime_field_overflow": "22008",
        "division_by_zero": "22012",
        "error_in_assignment": "22005",
        "escape_character_conflict": "2200B",
        "indicator_overflow": "22022",
        "interval_field_overflow": "22015",
        "invalid_argument_for_logarithm": "2201E",
        "invalid_argument_for_ntile_function": "22014",
        "invalid_argument_for_nth_value_function": "22016",
        "invalid_argument_for_power_function": "2201F",
        "invalid_argument_for_width_bucket_function": "2201G",
        "invalid_character_value_for_cast": "22018",
        "invalid_datetime_format": "22007",
        "invalid_escape_character": "22019",
        "invalid_escape_octet": "2200D",
        "invalid_escape_sequence": "22025",
        "nonstandard_use_of_escape_character": "22P06",
        "invalid_indicator_parameter_value": "22010",
        "invalid_parameter_value": "22023",
        "invalid_preceding_or_following_size": "22013",
        "invalid_regular_expression": "2201B",
        "invalid_row_count_in_limit_clause": "2201W",
        "invalid_row_count_in_result_offset_clause": "2201X",
        "invalid_tablesample_argument": "2202H",
        "invalid_tablesample_repeat": "2202G",
        "invalid_time_zone_displacement_value": "22009",
        "invalid_use_of_escape_character": "2200C",
        "most_specific_type_mismatch": "2200G",
        "null_value_not_allowed": "22004",
        "null_value_no_indicator_parameter": "22002",
        "numeric_value_out_of_range": "22003",
        "sequence_generator_limit_exceeded": "2200H",
        "string_data_length_mismatch": "22026",
        "string_data_right_truncation": "22001",
        "substring_error": "22011",
        "trim_error": "22027",
        "unterminated_c_string": "22024",
        "zero_length_character_string": "2200F",
        "floating_point_exception": "22P01",
        "invalid_text_representation": "22P02",
        "invalid_binary_representation": "22P03",
        "bad_copy_file_format": "22P04",
        "untranslatable_character": "22P05",
        "not_an_xml_document": "2200L",
        "invalid_xml_document": "2200M",
        "invalid_xml_content": "2200N",
        "invalid_xml_comment": "2200S",
        "invalid_xml_processing_instruction": "2200T",
        "duplicate_json_object_key_value": "22030",
        "invalid_argument_for_sql_json_datetime_function": "22031",
        "invalid_json_text": "22032",
        "invalid_sql_json_subscript": "22033",
        "more_than_one_sql_json_item": "22034",
        "no_sql_json_item": "22035",
        "non_numeric_sql_json_item": "22036",
        "non_unique_keys_in_a_json_object": "22037",
        "singleton_sql_json_item_required": "22038",
        "sql_json_array_not_found": "22039",
        "sql_json_member_not_found": "2203A",
        "sql_json_number_not_found": "2203B",
        "sql_json_object_not_found": "2203C",
        "too_many_json_array_elements": "2203D",
        "too_many_json_object_members": "2203E",
        "sql_json_scalar_required": "2203F",
        "sql_json_item_cannot_be_cast_to_target_type": "2203G"
    },
    "CLASS_23": {
        "integrity_constraint_violation": "23000",
        "restrict_violation": "23001",
        "not_null_violation": "23502",
        "foreign_key_violation": "23503",
        "unique_violation": "23505",
        "check_violation": "23514",
        "exclusion_violation": "23P01"},
    "CLASS_24": {
        "invalid_cursor_state": "24000"},
    "CLASS_25": {
        "invalid_transaction_state": "25000",
        "active_sql_transaction": "25001",
        "branch_transaction_already_active": "25002",
        "held_cursor_requires_same_isolation_level": "25008",
        "inappropriate_access_mode_for_branch_transaction": "25003",
        "inappropriate_isolation_level_for_branch_transaction": "25004",
        "no_active_sql_transaction_for_branch_transaction": "25005",
        "read_only_sql_transaction": "25006",
        "schema_and_data_statement_mixing_not_supported": "25007",
        "no_active_sql_transaction": "25P01",
        "in_failed_sql_transaction": "25P02",
        "idle_in_transaction_session_timeout": "25P03"},
    "CLASS_26": {
        "invalid_sql_statement_name": "26000"},
    "CLASS_27": {
        "triggered_data_change_violation": "27000"},
    "CLASS_28": {
        "invalid_authorization_specification": "28000",
        "invalid_password": "28P01",
        "dependent_privilege_descriptors_still_exist": "2B000",
        "dependent_objects_still_exist": "2BP01"},
    "CLASS_2D": {

        "invalid_transaction_termination": "2D000"},
    "CLASS_2F": {

        "sql_routine_exception": "2F000",
        "function_executed_no_return_statement": "2F005",
        "modifying_sql_data_not_permitted": "2F002",
        "prohibited_sql_statement_attempted": "2F003",
        "reading_sql_data_not_permitted": "2F004"},
    "CLASS_34": {
        "invalid_cursor_name": "34000"},
    "CLASS_38": {
        "external_routine_exception": "38000",
        "containing_sql_not_permitted": "38001",
        "modifying_sql_data_not_permitted": "38002",
        "prohibited_sql_statement_attempted": "38003",
        "reading_sql_data_not_permitted": "38004"},
    "CLASS_39": {
        "external_routine_invocation_exception": "39000",
        "invalid_sqlstate_returned": "39001",
        "null_value_not_allowed": "39004",
        "trigger_protocol_violated": "39P01",
        "srf_protocol_violated": "39P02",
        "event_trigger_protocol_violated": "39P03"},
    "CLASS_3B": {
        "savepoint_exception": "3B000",
        "invalid_savepoint_specification": "3B001"
    },
    "CLASS_3D": {
        "invalid_catalog_name": "3D000"},
    "CLASS_3F": {
        "invalid_schema_name": "3F000"},
    "CLASS_40": {
        "transaction_rollback": "40000",
        "transaction_integrity_constraint_violation": "40002",
        "serialization_failure": "40001",
        "statement_completion_unknown": "40003",
        "deadlock_detected": "40P01"},
    "CLASS_42": {  # Syntax Errors
        "syntax_error_or_access_rule_violation": "42000",
        # DEFAULTS to here:
        "syntax_error": "42601",
        "insufficient_privilege": "42501",
        "cannot_coerce": "42846",
        "grouping_error": "42803",
        "windowing_error": "42P20",
        "invalid_recursion": "42P19",
        "invalid_foreign_key": "42830",
        "invalid_name": "42602",
        "name_too_long": "42622",
        "reserved_name": "42939",
        "datatype_mismatch": "42804",
        "indeterminate_datatype": "42P18",
        "collation_mismatch": "42P21",
        "indeterminate_collation": "42P22",
        "wrong_object_type": "42809",
        "generated_always": "428C9",
        "undefined_column": "42703",
        "undefined_function": "42883",
        "undefined_table": "42P01",
        "undefined_parameter": "42P02",
        "undefined_object": "42704",
        "duplicate_column": "42701",
        "duplicate_cursor": "42P03",
        "duplicate_database": "42P04",
        "duplicate_function": "42723",
        "duplicate_prepared_statement": "42P05",
        "duplicate_schema": "42P06",
        "duplicate_table": "42P07",
        "duplicate_alias": "42712",
        "duplicate_object": "42710",
        "ambiguous_column": "42702",
        "ambiguous_function": "42725",
        "ambiguous_parameter": "42P08",
        "ambiguous_alias": "42P09",
        "invalid_column_reference": "42P10",
        "invalid_column_definition": "42611",
        "invalid_cursor_definition": "42P11",
        "invalid_database_definition": "42P12",
        "invalid_function_definition": "42P13",
        "invalid_prepared_statement_definition": "42P14",
        "invalid_schema_definition": "42P15",
        "invalid_table_definition": "42P16",
        "invalid_object_definition": "42P17"},

    "CLASS_44": {
        "with_check_option_violation": "44000"},

    "CLASS_53": {
        "insufficient_resources": "53000",
        "disk_full": "53100",
        "out_of_memory": "53200",
        "too_many_connections": "53300",
        "configuration_limit_exceeded": "53400", },

    "CLASS_54": {
        "program_limit_exceeded": "54000",
        "statement_too_complex": "54001",
        "too_many_columns": "54011",
        "too_many_arguments": "54023"},
    "CLASS_55": {
        "object_not_in_prerequisite_state": "55000",
        "object_in_use": "55006",
        "cant_change_runtime_param": "55P02",
        "lock_not_available": "55P03",
        "unsafe_new_enum_value_usage": "55P04"},
    "CLASS_57": {
        "operator_intervention": "57000",
        "query_canceled": "57014",
        "admin_shutdown": "57P01",
        "crash_shutdown": "57P02",
        "cannot_connect_now": "57P03",
        "database_dropped": "57P04",
        "idle_session_timeout": "57P05"},
    "CLASS_58": {
        "system_error": "58000",
        "io_error": "58030",
        "undefined_file": "58P01",
        "duplicate_file": "58P02"},
    "CLASS_72": {
        "snapshot_too_old": "72000"},
    "CLASS_F0": {
        "config_file_error": "F0000",
        "lock_file_exists": "F0001"},
    "CLASS_HV": {
        "fdw_error": "HV000",
        "fdw_column_name_not_found": "HV005",
        "fdw_dynamic_parameter_value_needed": "HV002",
        "fdw_function_sequence_error": "HV010",
        "fdw_inconsistent_descriptor_information": "HV021",
        "fdw_invalid_attribute_value": "HV024",
        "fdw_invalid_column_name": "HV007",
        "fdw_invalid_column_number": "HV008",
        "fdw_invalid_data_type": "HV004",
        "fdw_invalid_data_type_descriptors": "HV006",
        "fdw_invalid_descriptor_field_identifier": "HV091",
        "fdw_invalid_handle": "HV00B",
        "fdw_invalid_option_index": "HV00C",
        "fdw_invalid_option_name": "HV00D",
        "fdw_invalid_string_length_or_buffer_length": "HV090",
        "fdw_invalid_string_format": "HV00A",
        "fdw_invalid_use_of_null_pointer": "HV009",
        "fdw_too_many_handles": "HV014",
        "fdw_out_of_memory": "HV001",
        "fdw_no_schemas": "HV00P",
        "fdw_option_name_not_found": "HV00J",
        "fdw_reply_handle": "HV00K",
        "fdw_schema_not_found": "HV00Q",
        "fdw_table_not_found": "HV00R",
        "fdw_unable_to_create_execution": "HV00L",
        "fdw_unable_to_create_reply": "HV00M",
        "fdw_unable_to_establish_connection": "HV00N"},
    "CLASS_P0": {
        "plpgsql_error": "P0000",
        "raise_exception": "P0001",
        "no_data_found": "P0002",
        "too_many_rows": "P0003",
        "assert_failure": "P0004"},
    "CLASS_XX": {
        "internal_error": "XX000",
        "data_corrupted": "XX001",
        "index_corrupted": "XX002"
    }
}
POSTGRES_SYNTAX_ERROR_CODE = POSTGRES_ERROR_CODES["CLASS_42"]["syntax_error"]



from mindsdb.integrations.handlers.youtube_handler.youtube_tables import (
    YoutubeCommentsTable,
    YoutubeChannelsTable,
    YoutubeVideosTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities.log import get_log
from mindsdb_sql import parse_sql

from collections import OrderedDict
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from googleapiclient.discovery import build

logger = get_log("integrations.youtube_handler")


class YoutubeHandler(APIHandler):
    """Youtube handler implementation"""

    def __init__(self, name=None, **kwargs):
        """Initialize the Youtube handler.
        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})

        self.parser = parse_sql
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.connection = None
        self.is_connected = False

        youtube_video_comments_data = YoutubeCommentsTable(self)
        self._register_table("comments", youtube_video_comments_data)

        youtube_channel_data = YoutubeChannelsTable(self)
        self._register_table("channels", youtube_channel_data)

        youtube_video_data = YoutubeVideosTable(self)
        self._register_table("videos", youtube_video_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection

        youtube = build(
            "youtube", "v3", developerKey=self.connection_data["youtube_api_token"]
        )
        self.connection = youtube

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.
        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Youtube API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    youtube_access_token={
        "type": ARG_TYPE.STR,
        "description": "API Token for accessing Youtube Application API",
        "required": True,
        "label": "Youtube access token",
    }
)

connection_args_example = OrderedDict(youtube_api_token="<your-youtube-api-token>")



"""add_skills_table

Revision ID: 4c26ad04eeaa
Revises: b5bf593ba659
Create Date: 2023-08-31 17:04:26.898015

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4c26ad04eeaa'
down_revision = 'd44ab65a6a35'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'skills',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('type', sa.String(), nullable=False),
        sa.Column('params', sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('skills')



from textwrap import dedent
from collections import OrderedDict

from pandas import DataFrame

from hdbcli import dbapi
import sqlalchemy_hana.dialect as hana_dialect

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.utilities import log

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class HanaHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the SAP Hana statements.
    """

    name = 'hana'

    def __init__(self, name: str, connection_data: dict, **kwargs):
        super().__init__(name)

        self.dialect = 'hana'
        self.parser = parse_sql
        self.connection_data = connection_data
        self.renderer = SqlalchemyRender(hana_dialect.HANAHDBCLIDialect)

        self.address = self.connection_data.get('host')
        self.port = self.connection_data.get('port')
        self.user = self.connection_data.get('user')
        self.password = self.connection_data.get('password')
        self.autocommit = self.connection_data.get('autocommit', True)
        self.properties = self.connection_data.get('properties')
        self.currentSchema = self.connection_data.get('schema', 'CURRENTUSER')
        self.databaseName = self.connection_data.get('database')
        self.encrypt = self.connection_data.get('encrypt', False)
        self.sslHostNameInCertificate = self.connection_data.get('sslHostNameInCertificate')
        self.sslValidateCertificate = self.connection_data.get('sslValidateCertificate', False)
        self.sslCryptoProvider = self.connection_data.get('sslCryptoProvider')
        self.sslTrustStore = self.connection_data.get('sslTrustStore')
        self.sslKeyStore = self.connection_data.get('sslKeyStore')
        self.cseKeyStorePassword = self.connection_data.get('cseKeyStorePassword')
        self.sslSNIHostname = self.connection_data.get('sslSNIHostname')
        self.sslSNIRequest = self.connection_data.get('sslSNIRequest', True)
        self.siteType = self.connection_data.get('siteType')
        self.splitBatchCommands = self.connection_data.get('splitBatchCommands', True)
        self.routeDirectExecute = self.connection_data.get('routeDirectExecute', False)
        self.secondarySessionFallback = self.connection_data.get('secondarySessionFallback', True)

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Handles the connection to a SAP Hana database insance.
        """

        if self.is_connected is True:
            return self.connection

        connection = dbapi.connect(
            address=self.address,
            port=self.port,
            user=self.user,
            password=self.password,
            autocommit=self.autocommit,
            properties=self.properties,
            currentSchema=self.currentSchema,
            databaseName=self.databaseName,
            encrypt=self.encrypt,
            sslHostNameInCertificate=self.sslHostNameInCertificate,
            sslValidateCertificate=self.sslValidateCertificate,
            sslCryptoProvider=self.sslCryptoProvider,
            sslTrustStore=self.sslTrustStore,
            sslKeyStore=self.sslKeyStore,
            cseKeyStorePassword=self.cseKeyStorePassword,
            sslSNIHostname=self.sslSNIHostname,
            sslSNIRequest=self.sslSNIRequest,
            siteType=self.siteType,
            splitBatchCommands=self.splitBatchCommands,
            routeDirectExecute=self.routeDirectExecute,
            secondarySessionFallback=self.secondarySessionFallback
        )

        self.is_connected = True
        self.connection = connection
        return self.connection

    def disconnect(self):
        """
        Disconnects from the SAP HANA database
        """

        if self.is_connected is True:
            self.connection.close()
            self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the SAP HANA database
        :return: success status and error message if error occurs
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                cur.execute('SELECT 1 FROM SYS.DUMMY')
            response.success = True
        except dbapi.Error as e:
            log.logger.error(f'Error connecting to SAP HANA {self.address}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in SAP HANA
        :return: returns the records from the current recordset
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                if not cur.description:
                    response = Response(RESPONSE_TYPE.OK)
                else:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                connection.commit()
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.address}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """

        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all tables in SAP HANA in the current schema
        """

        return self.native_query(f"""
            SELECT SCHEMA_NAME,
                   TABLE_NAME,
                   TABLE_TYPE
            FROM
                SYS.TABLES
            WHERE IS_SYSTEM_TABLE = 'FALSE'  
              AND IS_USER_DEFINED_TYPE = 'FALSE'
              AND IS_TEMPORARY = 'FALSE'
        """)

    def get_columns(self, table_name: str) -> Response:
        """
        List all columns in a table in SAP HANA in the current schema
        :param table_name: the table name for which to list the columns
        :return: returns the columns in the table
        """

        return self.renderer.dialect.get_columns(table_name)


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The IP address/host name of the SAP HANA instance host.'
    },
    port={
        'type': ARG_TYPE.STR,
        'description': 'The port number of the SAP HANA instance.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the user name.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the password for the user.'
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': 'Sets the current schema, which is used for identifiers without a schema.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the name of the database to connect to. (Not used for SAP HANA Cloud)'
    },
    autocommit={
        'type': ARG_TYPE.BOOL,
        'description': 'Sets the autocommit mode for the connection.'
    },
    properties={
        'type': ARG_TYPE.STR,
        'description': 'Additional dictionary with special properties of the connection.'
    },
    encrypt={
        'type': ARG_TYPE.BOOL,
        'description': 'Enables or disables TLS encryption.'
    },
    sslHostNameInCertificate={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the host name used to verify server\'s identity.'
    },
    sslValidateCertificate={
        'type': ARG_TYPE.BOOL,
        'description': 'Specifies whether to validate the server\'s certificate.'
    },
    sslCryptoProvider={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the cryptographic library provider used for TLS communication.'
    },
    sslTrustStore={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the path to a trust store file that contains the server\'s public certificates.'
    },
    sslKeyStore={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the path to the keystore file that contains the client\'s identity.'
    },
    cseKeyStorePassword={
        'type': ARG_TYPE.STR,
        'description': 'Provides the password for the local key store.'
    },
    sslSNIHostname={
        'type': ARG_TYPE.STR,
        'description': dedent("""Specifies the name of the host that is attempting to connect at the start of
        the TLS handshaking process.""")
    },
    sslSNIRequest={
        'type': ARG_TYPE.BOOL,
        'description': 'Specifies whether SNI requests are enabled for TLS connections: TRUE/FALSE.'
    },
    siteType={
        'type': ARG_TYPE.STR,
        'description': dedent("""Specifies whether the connection is made to either the PRIMARY or SECONDARY
        site in an Active/Active (read enabled) system.""")
    },
    splitBatchCommands={
        'type': ARG_TYPE.BOOL,
        'description': 'Allows split and parallel execution of batch commands on partitioned tables.'
    },
    routeDirectExecute={
        'type': ARG_TYPE.BOOL,
        'description': dedent("""Converts direct execute into prepare and execute (routed execute) if the
        number of index servers is more than one and if statement routing is enabled.""")
    },
    secondarySessionFallback={
        'type': ARG_TYPE.BOOL,
        'description': dedent("""Forces the ongoing transaction on a non-anchor connection to fall back
        to the anchor/primary connection if this connection is dropped by the network or server.""")
    }
)

connection_args_example = OrderedDict(
    host='<uuid>.hana.trial-us10.hanacloud.ondemand.com',
    port=30013,
    user='DBADMIN',
    password='password',
    schema='MINDSDB',
)
